import os
import unittest
from collections import OrderedDict

import time
from time import sleep

from conans import REVISIONS, load
from conans.client.tools import environment_append
from conans.model.ref import ConanFileReference, PackageReference
from conans import DEFAULT_REVISION_V1
from conans.test.utils.tools import TestClient, TestServer, create_local_git_repo, \
    NO_SETTINGS_PACKAGE_ID


@unittest.skipUnless(TestClient().revisions,
                     "The test needs revisions activated, set CONAN_CLIENT_REVISIONS_ENABLED=1")
class RevisionsTest(unittest.TestCase):

    def setUp(self):
        self.servers = OrderedDict()
        self.users = {}
        for i in range(3):
            self.servers["remote%d" % i] = TestServer(server_capabilities=[REVISIONS])
            self.users["remote%d" % i] = [("lasote", "mypass")]

        self.servers["remote_norevisions"] = TestServer(server_capabilities=[])
        self.users["remote_norevisions"] = [("lasote", "mypass")]
        self.client = TestClient(servers=self.servers, users=self.users)
        self.conanfile = '''
from conans import ConanFile

class HelloConan(ConanFile):
    def build(self):
        self.output.warn("Revision 1")        
'''
        self.ref = ConanFileReference.loads("lib/1.0@lasote/testing")

    def _create_and_upload(self, conanfile, reference, remote=None, args=""):
        remote = remote or "remote0"
        self.client.save({"conanfile.py": conanfile})
        self.client.run("create . %s %s" % (str(reference), args))
        self.client.run("upload %s -c --all -r %s" % (str(reference), remote))

    def test_revisions_recipes_without_scm(self):

        self._create_and_upload(self.conanfile, self.ref)
        rev = self.servers["remote0"].paths.get_last_revision(self.ref).revision
        self.assertEquals(rev, "149570a812b46d87c7dfa6408809b370")

        # Create a new revision and upload
        conanfile = self.conanfile.replace("Revision 1", "Revision 2")
        self._create_and_upload(conanfile, self.ref)

        # Remove local and install latest
        self.client.run("remove %s -f" % str(self.ref))
        self.client.run("install %s --build" % str(self.ref))
        self.assertIn("Revision 2", self.client.out)

        # Remove local and install first
        self.client.run("remove %s -f" % str(self.ref))
        self.client.run("install %s#149570a812b46d87c7dfa6408809b370 --build" % str(self.ref))
        self.assertIn("Revision 1", self.client.out)

    def test_revision_with_fixed_scm(self):

        conanfile = '''
from conans import ConanFile

class HelloConan(ConanFile):
    scm = {"type": "git",
           "url": "auto",
           "revision": "fixed_revision"}       
'''
        path, commit = create_local_git_repo({"myfile": "contents",
                                              "conanfile.py": conanfile}, branch="my_release")
        self.client.current_folder = path
        self.client.runner('git remote add origin https://myrepo.com.git', cwd=path)
        self.client.save({"conanfile.py": conanfile})
        self.client.run("export . %s " % str(self.ref))
        rev = self.client.get_revision(self.ref)
        self.assertNotEqual(rev, "fixed_revision")

    def test_revisions_packages_download(self):
        conanfile = '''
import os
from conans import ConanFile, tools

class HelloConan(ConanFile):
    
    def build(self):
        tools.save("myfile.txt", os.getenv("PACKAGE_CONTENTS"))
        
    def package(self):
        self.copy("*")
'''
        with environment_append({"PACKAGE_CONTENTS": "1"}):
            self._create_and_upload(conanfile, self.ref)
        rev = self.servers["remote0"].paths.get_last_revision(self.ref).revision
        self.assertEquals(rev, "202f9ce41808083a0f0c0d071fb5f398")

        self.ref = self.ref.copy_with_rev(rev)
        p_ref = PackageReference(self.ref, "5ab84d6acfe1f23c4fae0ab88f26e3a396351ac9")
        pkg_rev = self.servers["remote0"].paths.get_last_package_revision(p_ref).revision
        self.assertEquals(pkg_rev, "15ab113a16e2ac8c9ecffb4ba48306b2")

        # Create new package revision for the same recipe
        with environment_append({"PACKAGE_CONTENTS": "2"}):
            self._create_and_upload(conanfile, self.ref.copy_clear_rev())
        pkg_rev = self.servers["remote0"].paths.get_last_package_revision(p_ref).revision
        self.assertEquals(pkg_rev, "8e54c6ea967722f2f9bdcbacb21792f5")

        # Delete all from local
        self.client.run("remove %s -f" % str(self.ref.copy_clear_rev()))

        # Download specifying recipe with revisions and package with revisions
        self.client.run("download %s -p 5ab84d6acfe1f23c4fae0ab88f26e3a396351ac9#"
                        "8e54c6ea967722f2f9bdcbacb21792f5" % self.ref.full_repr())

        contents = load(os.path.join(self.client.paths.package(p_ref), "myfile.txt"))
        self.assertEquals(contents, "2")

        # Download previous package revision
        self.client.run("download %s -p 5ab84d6acfe1f23c4fae0ab88f26e3a396351ac9#"
                        "15ab113a16e2ac8c9ecffb4ba48306b2" % self.ref.full_repr())
        contents = load(os.path.join(self.client.paths.package(p_ref), "myfile.txt"))
        self.assertEquals(contents, "1")

        # Specify a package revision without a recipe revision
        error = self.client.run("download %s -p 5ab84d6acfe1f23c4fae0ab88f26e3a396351ac9#"
                                "15ab113a16e2ac8c9ecffb4ba48306b2" % str(self.ref),
                                ignore_error=True)
        self.assertTrue(error)
        self.assertIn("It is needed to specify the recipe revision if "
                      "you specify a package revision", self.client.out)

    def test_search_with_revision(self):
        conanfile = '''
from conans import ConanFile

class HelloConan(ConanFile):
    settings = "os"
    def build(self):
        self.output.warn("Revision 1")        
'''
        self._create_and_upload(conanfile, self.ref, args="-s os=Linux")
        rev1 = self.servers["remote0"].paths.get_last_revision(self.ref).revision
        self._create_and_upload(conanfile.replace('"os"', '"arch"'), self.ref, args="-s arch=x86")
        rev2 = self.servers["remote0"].paths.get_last_revision(self.ref).revision
        self.assertNotEqual(rev1, rev2)

        # Search every package in local cache (we get both binary packages)
        self.client.run("search lib/1.0@lasote/testing")
        self.assertIn("a363db07e8420d258dca5a64aad6a5b8ecbbdd66", self.client.out)
        self.assertIn("cb054d0b3e1ca595dc66bc2339d40f1f8f04ab31", self.client.out)

        # Search only revision 1 in local cache
        self.client.run("search lib/1.0@lasote/testing#%s" % rev1)
        self.assertNotIn("a363db07e8420d258dca5a64aad6a5b8ecbbdd66", self.client.out)
        self.assertIn("cb054d0b3e1ca595dc66bc2339d40f1f8f04ab31", self.client.out)

        # Search only revision 2 in local cache
        self.client.run("search lib/1.0@lasote/testing#%s" % rev2)
        self.assertIn("a363db07e8420d258dca5a64aad6a5b8ecbbdd66", self.client.out)
        self.assertNotIn("cb054d0b3e1ca595dc66bc2339d40f1f8f04ab31", self.client.out)

        # Search all in remote (will give us the latest)
        self.client.run("search lib/1.0@lasote/testing -r remote0")
        self.assertIn("a363db07e8420d258dca5a64aad6a5b8ecbbdd66", self.client.out)
        self.assertNotIn("cb054d0b3e1ca595dc66bc2339d40f1f8f04ab31", self.client.out)

        # Search rev1 in remote
        self.client.run("search lib/1.0@lasote/testing#%s -r remote0" % rev1)
        self.assertNotIn("a363db07e8420d258dca5a64aad6a5b8ecbbdd66", self.client.out)
        self.assertIn("cb054d0b3e1ca595dc66bc2339d40f1f8f04ab31", self.client.out)

        # Search rev2 in remote
        self.client.run("search lib/1.0@lasote/testing#%s -r remote0" % rev2)
        self.assertIn("a363db07e8420d258dca5a64aad6a5b8ecbbdd66", self.client.out)
        self.assertNotIn("cb054d0b3e1ca595dc66bc2339d40f1f8f04ab31", self.client.out)

    def test_with_scm(self):
        conanfile = '''
from conans import ConanFile

class HelloConan(ConanFile):
    scm = {
        "revision": "auto",
        "url": "auto",
        "type": "git"
    }
    def build(self):
        self.output.warn("Revision 1")        
'''
        path, commit = create_local_git_repo({"myfile": "contents",
                                              "conanfile.py": conanfile}, branch="my_release")
        self.client.runner('git remote add origin https://myrepo.com.git', cwd=path)

        self.client.current_folder = path
        self.client.run("create . %s" % str(self.ref))
        self.client.run("upload %s -c --all -r remote0" % str(self.ref))
        rev_server = self.servers["remote0"].paths.get_last_revision(self.ref)
        self.assertEqual(commit, rev_server.revision)

        self.client.run("remove %s -f" % str(self.ref))
        self.client.run("install %s#%s" % (str(self.ref), rev_server.revision))
        self.assertIn("Package installed 5ab84d6acfe1f23c4fae0ab88f26e3a396351ac9", self.client.out)

        self.client.run("remove %s -f" % str(self.ref))
        self.client.run("install %s" % str(self.ref))
        self.assertIn("Package installed 5ab84d6acfe1f23c4fae0ab88f26e3a396351ac9", self.client.out)

    def test_info(self):
        conanfile = '''
from conans import ConanFile

class HelloConan(ConanFile):
    settings = "os"
    def build(self):
        self.output.warn("Revision 1")        
'''
        self._create_and_upload(conanfile, self.ref, args="-s os=Linux")
        self.client.run("info %s" % str(self.ref))
        self.assertIn("Revision: c5485544fd84cf85e45cc742feb8b34c", self.client.out)

        self.client.remote_registry.refs.remove(self.ref)
        # Upload to a non-revisions server, the revision should be always there in the registry
        self._create_and_upload(conanfile, self.ref, args="-s os=Linux", remote="remote_norevisions")
        self.client.run("info %s" % str(self.ref))
        self.assertIn("Revision: c5485544fd84cf85e45cc742feb8b34c", self.client.out)

    def test_update_recipe(self):
        self.client.save({"conanfile.py": self.conanfile})
        self.client.run("create . %s" % str(self.ref))
        self.client.run("upload %s -c --all -r remote0" % str(self.ref))
        self.client.run("remote list_ref")

        pref = PackageReference.loads("lib/1.0@lasote/testing:"
                                      "5ab84d6acfe1f23c4fae0ab88f26e3a396351ac9")
        self.assertIn("%s: remote0" % str(self.ref), self.client.out)
        self.client.run("remote list_pref %s" % str(self.ref))
        self.assertIn("%s: remote0" % str(pref), self.client.out)

        rev = self.client.get_revision(self.ref)
        self.assertEquals(rev, "149570a812b46d87c7dfa6408809b370")

        sleep(1)
        client2 = TestClient(servers=self.servers, users=self.users)
        conanfile2 = self.conanfile + "# Holaaa "
        client2.save({"conanfile.py": conanfile2})
        client2.run("create . %s" % str(self.ref))
        client2.run("upload %s -c --all -r remote0" % str(self.ref))
        client2.run("remote list_pref %s" % str(self.ref))

        rev = client2.get_revision(self.ref)
        self.assertEquals(rev, "621568e8053761d685dcf1bfbe3b3f10")

        # install of the client1 (no-update)
        self.client.run("install %s" % str(self.ref))
        self.assertIn("lib/1.0@lasote/testing from 'remote0' - Cache", self.client.out)
        self.assertIn("%s - Cache" % str(pref), self.client.out)
        self.client.run("remote list_pref %s" % str(self.ref))
        self.assertIn("%s: remote0" % str(pref), self.client.out)

        rev = self.client.get_revision(self.ref)
        self.assertEquals(rev, "149570a812b46d87c7dfa6408809b370")

        # install with update
        self.client.run("install %s --update" % str(self.ref))
        rev = self.client.get_revision(self.ref)
        self.assertEquals(rev, "621568e8053761d685dcf1bfbe3b3f10")
        self.assertNotIn("%s from 'remote0' - Newer" % str(self.ref), self.client.out)
        self.assertIn("Outdated package! The package doesn't belong to the installed recipe "
                      "revision:", self.client.out)
        self.assertIn("%s from 'remote0' - Updated" % str(self.ref), self.client.out)
        self.assertIn("%s - Update" % str(pref), self.client.out)

    def test_registry_revision_updated(self):
        self.client.save({"conanfile.py": self.conanfile})
        self.client.run("create . %s" % str(self.ref))
        self.client.run("upload %s -c --all -r remote0" % str(self.ref))
        rev = self.client.get_revision(self.ref)
        remote_rev = self.servers["remote0"].paths.get_last_revision(self.ref)
        self.assertEquals(remote_rev.revision, rev)

        self.client.save({"conanfile.py": self.conanfile + " "})
        self.client.run("create . %s" % str(self.ref))
        self.client.run("upload %s -c --all -r remote0" % str(self.ref))
        rev2 = self.client.get_revision(self.ref)
        remote_rev2 = self.servers["remote0"].paths.get_last_revision(self.ref)
        self.assertEquals(remote_rev2.revision, rev2)
        self.assertNotEquals(rev, rev2)

    def test_update_package(self):
        conanfile = '''
import time
import os
from conans import ConanFile, tools

class HelloConan(ConanFile):
    def package(self):
        tools.save(os.path.join(self.package_folder, "file.txt"), str(time.time()))
'''

        self.client.save({"conanfile.py": conanfile})
        self.client.run("create . %s" % str(self.ref))
        self.client.run("upload %s -c --all -r remote0" % str(self.ref))

        # Same recipe, but will generate different package
        sleep(1)
        client2 = TestClient(servers=self.servers, users=self.users)
        client2.save({"conanfile.py": conanfile})
        client2.run("create . %s" % str(self.ref))
        client2.run("upload %s -c --all -r remote0" % str(self.ref))

        # install of the client1 (no-update)
        self.client.run("install %s" % str(self.ref))
        self.assertIn("lib/1.0@lasote/testing from 'remote0' - Cache", self.client.out)
        self.assertIn("lib/1.0@lasote/testing:5ab84d6acfe1f23c4fae0ab88f26e3a396351ac9 - Cache",
                      self.client.out)

        # install with update
        self.client.run("install %s --update" % str(self.ref))
        self.assertNotIn("The current binary package doesn't belong to the current recipe revision:",
                         self.client.out)
        self.assertIn("Current package is older than remote upstream one", self.client.out)
        self.assertIn("lib/1.0@lasote/testing from 'remote0' - Cache", self.client.out)
        self.assertIn("lib/1.0@lasote/testing:5ab84d6acfe1f23c4fae0ab88f26e3a396351ac9 - Update",
                      self.client.out)

    def test_upload_outdated_packages(self):
        conanfile = '''
from conans import ConanFile

class HelloConan(ConanFile):
    settings = "os"      
'''
        self.client.save({"conanfile.py": conanfile})
        self.client.run("create . %s -s os=Windows" % str(self.ref))
        self.client.run("upload %s -c --all -r remote0" % str(self.ref))

        conanfile2 = self.conanfile + " "
        self.client.save({"conanfile.py": conanfile2})
        self.client.run("create . %s -s os=Linux" % str(self.ref))
        self.client.run("upload %s -c --all -r remote0" % str(self.ref))
        self.assertIn("Skipping package '3475bd55b91ae904ac96fde0f106a136ab951a5e'",
                      self.client.out)
        self.assertIn("Uploading package 1/1", self.client.out)

    def test_conflict_with_different_revisions_but_same_ref(self):
        """Diamond requiring different revisions in the reference"""
        conanfile = '''
from conans import ConanFile

class HelloConan(ConanFile): 
    pass
'''
        ref_a = ConanFileReference.loads("libA/1.0@lasote/testing")
        self._create_and_upload(conanfile, ref_a)
        rev = self.servers["remote0"].paths.get_last_revision(ref_a).revision
        self._create_and_upload(conanfile + " ", ref_a)  # New revision
        rev2 = self.servers["remote0"].paths.get_last_revision(ref_a).revision

        ref_b = ConanFileReference.loads("libB/1.0@lasote/testing")
        req = "%s#%s" % (ref_a, rev)
        self._create_and_upload(conanfile.replace("pass", 'requires = "%s"' % req), ref_b)

        ref_c = ConanFileReference.loads("libC/1.0@lasote/testing")
        req = "%s#%s" % (ref_a, rev2)
        self._create_and_upload(conanfile.replace("pass", 'requires = "%s"' % req), ref_c)

        ref_d = ConanFileReference.loads("libD/1.0@lasote/testing")
        repl = 'requires = "%s", "%s"' % (str(ref_c), str(ref_b))
        self.client.save({"conanfile.py": conanfile.replace("pass", repl)})
        error = self.client.run("create . %s" % str(ref_d), ignore_error=True)

        self.assertTrue(error)
        self.assertIn("Different revisions of libA/1.0@lasote/testing "
                      "has been requested", self.client.out)

        self.client.run('remove "*" -f')
        self.client.run("create . %s" % str(ref_d), ignore_error=True)
        error = self.client.run("create . %s" % str(ref_d), ignore_error=True)
        self.assertTrue(error)
        self.assertIn("Different revisions of libA/1.0@lasote/testing "
                      "has been requested", self.client.out)

    def test_upload_not_overwrite(self):

        self.client.save({"conanfile.py": self.conanfile})
        self.client.run("create . %s" % str(self.ref))
        self.client.run("upload %s -c --all --no-overwrite" % str(self.ref))
        self.client.save({"conanfile.py": self.conanfile + " "})
        self.client.run("create . %s" % str(self.ref))
        self.client.run("upload %s -c --all --no-overwrite" % str(self.ref))

    def test_export_cleans_revision_in_registy(self):
        self.client.save({"conanfile.py": self.conanfile})
        self.client.run("create . %s" % str(self.ref))
        self.client.run("upload %s -c --all --no-overwrite" % str(self.ref))

        # No changes
        self.client.save({"conanfile.py": self.conanfile})
        self.client.run("export . %s" % str(self.ref))
        cur_rev = self.client.get_revision(self.ref)
        self.assertIsNotNone(cur_rev)

        # Export new recipe, the revision is not cleared but changed
        self.client.save({"conanfile.py": self.conanfile + " "})
        self.client.run("export . %s" % str(self.ref))
        new_rev = self.client.get_revision(self.ref)
        self.assertIsNotNone(new_rev)
        self.assertNotEqual(cur_rev, new_rev)

    def test_alias_with_revisions(self):

        self.client.save({"conanfile.py": self.conanfile})
        self.client.run("create . %s" % str(self.ref))
        self.client.run("upload %s -c --all -r remote0" % str(self.ref))
        rev = self.client.get_revision(self.ref)
        full_ref = self.ref.copy_with_rev(rev)

        self.client.save({"conanfile.py": self.conanfile.replace("Revision 1", "Revision 2")})
        self.client.run("create . %s" % str(self.ref))
        self.client.run("upload %s -c --all -r remote0" % str(self.ref))

        alias = """from conans import ConanFile

class AliasConanfile(ConanFile):
    alias = "%s"
""" % full_ref.full_repr()

        self.client.save({"conanfile.py": alias})
        self.client.run("export . lib/snap@lasote/testing")
        # As we requested a different revision it will install the correct one
        self.client.run("install lib/snap@lasote/testing --build")
        self.assertIn("Revision 1", self.client.out)

    def test_recipe_revision_delete_all(self):
        self.client.save({"conanfile.py": self.conanfile})
        self.client.run("create . %s" % str(self.ref))
        self.client.run("upload %s -c --all -r remote0" % str(self.ref))

        self.client.save({"conanfile.py": self.conanfile.replace("Revision 1", "Revision 2")})
        self.client.run("create . %s" % str(self.ref))
        self.client.run("upload %s -c --all -r remote0" % str(self.ref))
        self.client.run("remove %s -r remote0 -f" % str(self.ref))

        last_rev = self.servers["remote0"].paths.get_last_revision(self.ref)
        self.assertIsNone(last_rev)

    def test_recipe_revision_delete_one(self):
        # Upload revision1
        self.client.save({"conanfile.py": self.conanfile})
        self.client.run("create . %s" % str(self.ref))
        self.client.run("upload %s -c --all -r remote0" % str(self.ref))
        rev = self.client.get_revision(self.ref)
        remote_rev = self.servers["remote0"].paths.get_last_revision(self.ref)
        self.assertEquals(remote_rev.revision, rev)

        # Upload revision2
        self.client.save({"conanfile.py": self.conanfile.replace("Revision 1", "Revision 2")})
        self.client.run("create . %s" % str(self.ref))
        self.client.run("upload %s -c --all -r remote0" % str(self.ref))
        rev2 = self.client.get_revision(self.ref)
        remote_rev = self.servers["remote0"].paths.get_last_revision(self.ref)
        self.assertEquals(remote_rev.revision, rev2)

        # Remove revision2
        self.client.run("remove %s#%s -r remote0 -f" % (str(self.ref), rev2))
        remote_rev = self.servers["remote0"].paths.get_last_revision(self.ref)
        self.assertEquals(remote_rev.revision, rev)

        # Upload revision3
        self.client.save({"conanfile.py": self.conanfile.replace("Revision 1", "Revision 3")})
        self.client.run("create . %s" % str(self.ref))
        self.client.run("upload %s -c --all -r remote0" % str(self.ref))
        rev3 = self.client.get_revision(self.ref)
        self.assertNotEquals(rev3, rev2)
        remote_rev = self.servers["remote0"].paths.get_last_revision(self.ref)
        self.assertEquals(remote_rev.revision, rev3)

        # Remove revision 3, remote is rev1
        self.client.run("remove %s#%s -r remote0 -f" % (str(self.ref), rev3))
        remote_rev = self.servers["remote0"].paths.get_last_revision(self.ref)
        self.assertEquals(remote_rev.revision, rev)

    def test_remote_search(self):
        conanfile = '''
import time
import os
from conans import ConanFile, tools

class HelloConan(ConanFile):
    settings = "os"
    def package(self):
        tools.save(os.path.join(self.package_folder, "file.txt"), str(time.time()))
'''
        self.client.save({"conanfile.py": conanfile})
        self.client.run("create . %s -s os=Windows" % str(self.ref))
        self.client.run("upload %s -c --all -r remote0" % str(self.ref))
        rev1 = self.servers["remote0"].paths.get_last_revision(self.ref).revision
        # Create other recipe revision with a different binary package
        self.client.save({"conanfile.py": conanfile + " "})
        self.client.run("create . %s -s os=Linux" % str(self.ref))
        self.client.run("upload %s -c --all -r remote0" % str(self.ref))
        rev2 = self.servers["remote0"].paths.get_last_revision(self.ref).revision
        # Search with revision1 (only win package available)
        self.client.run("search %s#%s -r remote0" % (str(self.ref), rev1))
        self.assertIn("Existing recipe in remote 'remote0'", self.client.out)
        self.assertIn("os: Windows", self.client.out)
        self.assertNotIn("os: Linux", self.client.out)
        # Search with revision2 (only linux package available)
        self.client.run("search %s#%s -r remote0" % (str(self.ref), rev2))
        self.assertIn("Existing recipe in remote 'remote0'", self.client.out)
        self.assertIn("os: Linux", self.client.out)
        self.assertNotIn("os: Windows", self.client.out)

    def _upload_two_revisions(self, ref, different_binary=False):
            conanfile = '''
import os
import uuid
from conans import ConanFile, tools

class HelloConan(ConanFile): 
    pass
'''
            if different_binary:  # Same pid but different binary hash
                conanfile += """
    def package(self):
        tools.save(os.path.join(self.package_folder, "file"), str(uuid.uuid4()))
"""
            self._create_and_upload(conanfile, ref)
            _rev1 = self.client.get_revision(ref)
            self._create_and_upload(conanfile + "\n", ref)
            _rev2 = self.client.get_revision(ref)
            return _rev1, _rev2

    def test_remove_recipe_v2_test(self):

        ref = ConanFileReference.loads("lib/1.0@lasote/testing")
        rev1, rev2 = self._upload_two_revisions(ref)

        # If I specify rev2 only rev2 is removed
        self.client.run("remove %s#%s -f -r remote0" % (str(ref), rev2))
        latestrev = self.servers["remote0"].paths.get_last_revision(ref).revision
        self.assertEquals(latestrev, rev1)

        self._upload_two_revisions(ref)
        # If I don't specify revision all the revisions are removed
        self.client.run("remove %s -f -r remote0" % str(ref))
        latestrev = self.servers["remote0"].paths.get_last_revision(ref)
        self.assertIsNone(latestrev)

    def test_remove_packages_v2_test(self):
        # If I don't specify a recipe revision it will remove all the packages from all recipe
        # revisions
        pid = "5ab84d6acfe1f23c4fae0ab88f26e3a396351ac9"
        ref = ConanFileReference.loads("lib/1.0@lasote/testing")
        rev1, rev2 = self._upload_two_revisions(ref)
        full_pr1 = PackageReference(ref.copy_with_rev(rev1), pid)
        full_pr2 = PackageReference(ref.copy_with_rev(rev2), pid)
        path1 = self.servers["remote0"].paths.package(full_pr1)
        path2 = self.servers["remote0"].paths.package(full_pr2)
        self.assertTrue(os.path.exists(path1))
        self.assertTrue(os.path.exists(path2))

        self.client.run("remove %s -f -r remote0 -p %s" % (str(ref), pid))
        self.assertFalse(os.path.exists(path1))
        self.assertFalse(os.path.exists(path2))

        # Now specify the recipe revision
        self._upload_two_revisions(ref)
        self.client.run("remove %s#%s -f -r remote0 -p %s" % (str(ref), rev1, pid))
        self.assertFalse(os.path.exists(path1))
        self.assertTrue(os.path.exists(path2))

        # Now generate different package revisions also
        rev1, rev2 = self._upload_two_revisions(ref, different_binary=True)
        full_pr1 = PackageReference(ref.copy_with_rev(rev1), pid)
        prevs = [el.revision for el in self.servers["remote0"].paths.get_package_revisions(full_pr1)]
        self.assertEquals(len(prevs), 1)
        self._upload_two_revisions(ref, different_binary=True)
        prevs = [el.revision for el in self.servers["remote0"].paths.get_package_revisions(full_pr1)]
        self.assertEquals(len(prevs), 2)

        # Remove a concrete package reference
        self.client.run("remove %s#%s -f -r remote0 -p %s#%s" % (str(ref), rev1, pid, prevs[0]))
        prevs_now = [el.revision
                     for el in self.servers["remote0"].paths.get_package_revisions(full_pr1)]
        self.assertEquals(len(prevs_now), 1)
        self.assertEquals(prevs_now[0], prevs[1])

    def test_remove_all_revs_with_v1(self):
        # Check if v1 with several versions in the server will:
        # Remove all the revisions and packages if I remove by reference
        conanfile = '''
from conans import ConanFile

class HelloConan(ConanFile): 
    pass
'''
        ref = ConanFileReference.loads("lib/1.0@lasote/testing")
        # Upload three revisions of A and its packages
        self._create_and_upload(conanfile, ref)
        rev1 = self.client.get_revision(ref)

        self._create_and_upload(conanfile + "\n", ref)
        rev2 = self.client.get_revision(ref)

        self._create_and_upload(conanfile + "\n\n", ref)
        rev3 = self.client.get_revision(ref)

        self._create_and_upload(conanfile + "\n\n\n", ref)
        rev4 = self.client.get_revision(ref)

        self.assertNotEquals(rev1, rev2)
        self.assertNotEquals(rev2, rev3)
        self.assertNotEquals(rev1, rev3)

        # Remove all locally
        self.client.run("remove '*' -f")

        # First a pre-check, if I remove only the package for the rev4 the rest is there
        self.client.run("remove %s#%s -f -r remote0" % (ref, rev4))
        latestrev = self.servers["remote0"].paths.get_last_revision(ref).revision
        self.assertEquals(latestrev, rev3)
        self.client.run("install %s#%s" % (self.ref, rev3))

        # Create a v1 client and remove the pid
        client_no_rev = TestClient(block_v2=True, servers=self.servers, users=self.users)
        client_no_rev.run("remove %s -f -p 5ab84d6acfe1f23c4fae0ab88f26e3a396351ac9 "
                          "-r remote0" % str(ref))

        # Remove all locally again
        self.client.run("remove '*' -f")

        # Use the regular v2 client and try to install specifying revisions
        error = self.client.run("install %s#%s" % (self.ref, rev1), ignore_error=True)
        self.assertTrue(error)
        self.assertIn("Can't find a 'lib/1.0@lasote/testing' package", self.client.out)
        error = self.client.run("install %s#%s" % (self.ref, rev2), ignore_error=True)
        self.assertTrue(error)
        self.assertIn("Can't find a 'lib/1.0@lasote/testing' package", self.client.out)
        error = self.client.run("install %s#%s" % (self.ref, rev3), ignore_error=True)
        self.assertTrue(error)
        self.assertIn("Can't find a 'lib/1.0@lasote/testing' package", self.client.out)

    def test_v1_with_revisions_behavior(self):

        client_no_rev = TestClient(block_v2=True, servers=self.servers, users=self.users)
        conanfile = '''
from conans import ConanFile

class HelloConan(ConanFile):
    
    def build(self):
        self.output.warn("Hello")     
'''
        client_no_rev.save({"conanfile.py": conanfile})
        client_no_rev.run("create . %s" % str(self.ref))
        client_no_rev.run("upload %s -c --all -r remote0" % str(self.ref))

        rev1 = self.servers["remote0"].paths.get_last_revision(self.ref).revision
        self.assertEquals(rev1, DEFAULT_REVISION_V1)

        # An upload from the other client with revisions puts a new revision as latest
        self.client.save({"conanfile.py": conanfile.replace("Hello", "Bye")})
        self.client.run("create . %s" % str(self.ref))
        self.client.run("upload %s -c --all -r remote0" % str(self.ref))

        rev1 = self.servers["remote0"].paths.get_last_revision(self.ref).revision
        self.assertNotEquals(rev1, DEFAULT_REVISION_V1)

        client_no_rev.run('remove "*" -f')
        client_no_rev.run("install %s --build" % str(self.ref))
        # Client v1 never receives revision but 0
        self.assertEquals(client_no_rev.get_revision(self.ref), DEFAULT_REVISION_V1)
        self.assertIn("Bye", client_no_rev.out)

        # If client v1 uploads again the recipe it is the latest again, but with rev0
        client_no_rev.save({"conanfile.py": conanfile.replace("Hello", "Foo")})
        client_no_rev.run("create . %s" % str(self.ref))
        client_no_rev.run("upload %s -c --all -r remote0" % str(self.ref))

        rev1 = self.servers["remote0"].paths.get_last_revision(self.ref).revision
        self.assertEquals(rev1, DEFAULT_REVISION_V1)

        # Even for the client with revision, now the latest is 0
        self.client.run('remove "*" -f')
        self.client.run("install %s --build" % str(self.ref))
        self.assertIn("Foo", self.client.out)
        self.assertEquals(self.client.get_revision(self.ref), DEFAULT_REVISION_V1)

    def package_iterating_remote_same_recipe_revision_test(self):
        """If remote1 and remote2 has the same recipe revisions it will
        look for binaries iterating"""
        pid = "5ab84d6acfe1f23c4fae0ab88f26e3a396351ac9"
        # https://github.com/conan-io/conan/issues/3882
        conanfile = """from conans import ConanFile
class ConanFileToolsTest(ConanFile):
    pass
"""
        # Upload recipe + package to remote1 and remote2
        ref = "Hello/0.1@lasote/stable"
        self.client.save({"conanfile.py": conanfile})
        self.client.run("create . %s" % ref)
        self.client.run("upload %s -r=remote0 --all" % ref)
        self.client.run("upload %s -r=remote2 --all" % ref)

        # Remove only binary from remote1 and everything in local
        self.client.run("remove -f %s -p -r remote0" % ref)
        self.client.run('remove "*" -f')

        # Now install it from a client, it will find the binary in remote2
        # because the recipe revision is the same
        self.client.run("install %s" % ref)
        self.assertIn("Retrieving package %s from remote 'remote2' " % pid, self.client.out)

    def package_iterating_remote_different_recipe_revision_test(self):
        """If remote1 and remote2 has the same recipe revisions it wont
        look for binaries iterating"""
        # https://github.com/conan-io/conan/issues/3882
        conanfile = """from conans import ConanFile
class ConanFileToolsTest(ConanFile):
    pass
"""
        # Upload recipe + package to remote1 and remote2
        ref = "Hello/0.1@lasote/stable"
        self.client.save({"conanfile.py": conanfile})
        self.client.run("create . %s" % ref)
        self.client.run("upload %s -r=remote0 --all" % ref)

        self.client.save({"conanfile.py": conanfile + "\n"})
        self.client.run("create . %s" % ref)
        self.client.run("upload %s -r=remote2 --all" % ref)

        # Remove only binary from remote1 and everything in local
        self.client.run("remove -f %s -p -r remote0" % ref)
        self.client.run('remove "*" -f')

        # Now install it from a client, it won't find the binary in remote2
        # because the recipe revision is NOT the same
        error = self.client.run("install %s" % ref, ignore_error=True)
        self.assertTrue(error)
        self.assertIn("Can't find a 'Hello/0.1@lasote/stable' package", self.client.out)


@unittest.skipUnless(TestClient().revisions,
                     "The test needs revisions activated, set CONAN_CLIENT_REVISIONS_ENABLED=1")
class CompatibilityRevisionsTest(unittest.TestCase):
    """Testing non breaking behavior from v1 and v2 with compatibility mode"""

    def setUp(self):
        self.servers = OrderedDict()
        self.servers["remote0"] = TestServer(server_capabilities=[REVISIONS])
        self.users = {"remote0": [("lasote", "mypass")],
                      "remote_norevisions": [("lasote", "mypass")]}
        self.servers["remote_norevisions"] = TestServer(server_capabilities=[])
        self.client = TestClient(servers=self.servers, users=self.users)
        self.conanfile = '''
from conans import ConanFile

class HelloConan(ConanFile):
    def build(self):
        self.output.warn("Revision 1")        
'''
        self.ref = ConanFileReference.loads("lib/1.0@lasote/testing")

    def test_search_v1_iterate_remotes(self):
        """Two recipe revisions, first with 1 binary, second with 1 binary, search v1 and v2
        with compatibility, have to find both"""
        conanfile = """from conans import ConanFile
class ConanFileToolsTest(ConanFile):
    settings = "os"
"""
        # Upload recipe + package to remote0
        ref = "Hello/0.1@lasote/stable"
        self.client.save({"conanfile.py": conanfile})
        self.client.run("create . %s -s os=Windows" % ref)
        self.client.run("upload %s -r=remote0 --all" % ref)

        conanfile += " "
        self.client.save({"conanfile.py": conanfile})
        self.client.run("create . %s -s os=Linux" % ref)
        self.client.run("upload %s -r=remote0 --all" % ref)

        client_no_rev = TestClient(block_v2=True, servers=self.servers, users=self.users)
        client_no_rev.run("search %s -r remote0" % ref)

        self.assertIn("3475bd55b91ae904ac96fde0f106a136ab951a5e", client_no_rev.out)
        self.assertIn("cb054d0b3e1ca595dc66bc2339d40f1f8f04ab31", client_no_rev.out)

        client_compatibility = TestClient(block_v2=True, revisions=False,
                                          servers=self.servers, users=self.users)
        client_compatibility.run("search %s -r remote0" % ref)
        self.assertIn("3475bd55b91ae904ac96fde0f106a136ab951a5e", client_compatibility.out)
        self.assertIn("cb054d0b3e1ca595dc66bc2339d40f1f8f04ab31", client_compatibility.out)

        # And not repeated even if duplicated in remote
        self.client.run("create . %s -s os=Windows" % ref)
        self.client.run("upload %s -r=remote0 --all" % ref)

        client_no_rev.run("search %s -r remote0" % ref)

        self.assertEquals(str(client_no_rev.out).count("3475bd55b91ae904ac96fde0f106a136ab951a5e"),
                          1)
        self.assertEquals(str(client_no_rev.out).count("cb054d0b3e1ca595dc66bc2339d40f1f8f04ab31"),
                          1)

    def test_remove_packages_from_recipe_revision(self):
        """It shouldn't remove the packages for all recipe revisions but only for the specified
        if a recipe revision is specified"""
        conanfile = """from conans import ConanFile
class ConanFileToolsTest(ConanFile):
    pass
"""
        ref = ConanFileReference.loads("Hello/0.1@lasote/stable")

        # Upload recipe rev 1 + package to remote0
        self.client.save({"conanfile.py": conanfile})
        self.client.run("create . %s" % str(ref))
        self.client.run("upload %s -r=remote0 --all" % str(ref))
        rev1 = self.client.get_revision(ref)

        # Upload recipe rev 2 + package to remote0
        self.client.save({"conanfile.py": conanfile + "\n"})
        self.client.run("create . %s" % str(ref))
        self.client.run("upload %s -r=remote0 --all" % str(ref))
        rev2 = self.client.get_revision(ref)

        # Remove the binaries from the recipe rev1
        self.client.run("remove %s#%s -p -r=remote0 -f" % (str(ref), rev1))
        self.client.run("remove %s -f" % str(ref))

        # Try to install binaries from the rev1, it should fail
        error = self.client.run("install %s#%s" % (str(ref), rev1), ignore_error=True)
        self.assertTrue(error)
        self.assertIn("Missing prebuilt package for 'Hello/0.1@lasote/stable", self.client.out)

        # Try to install binaries from the rev2, it should succeed
        self.client.run("install %s#%s" % (str(ref), rev2))

        # Also try without specifying revision
        self.client.run("remove %s -f" % str(ref))
        self.client.run("install %s" % str(ref))

    def test_install_recipe_revision(self):
        """ Specifying the revision, it has to install that revision.
        """
        conanfile = """from conans import ConanFile
class ConanFileToolsTest(ConanFile):
    pass
"""
        ref = ConanFileReference.loads("Hello/0.1@lasote/stable")

        # Upload recipe rev 1 + package to remote0
        self.client.save({"conanfile.py": conanfile})
        self.client.run("create . %s" % str(ref))
        self.client.run("upload %s -r=remote0 --all" % str(ref))
        rev1 = self.client.get_revision(ref)

        # Upload recipe rev 2 + package to remote0
        self.client.save({"conanfile.py": conanfile + "\n#Comment for rev2"})
        self.client.run("create . %s" % str(ref))
        self.client.run("upload %s -r=remote0 --all" % str(ref))
        rev2 = self.client.get_revision(ref)

        self.assertNotEqual(rev1, rev2)

        # Remove all from local
        self.client.run("remove %s -f" % str(ref))

        # Try to install rev1 and not rev2
        self.client.run("install %s#%s" % (str(ref), rev1))
        conanfile_path = self.client.client_cache.conanfile(ref)
        contents = load(conanfile_path)
        self.assertNotIn("#Comment for rev2", contents)

        # Remove all from local
        self.client.run("remove %s -f" % str(ref))

        # Try to install rev2 and not rev1
        self.client.run("install %s#%s" % (str(ref), rev2))
        conanfile_path = self.client.client_cache.conanfile(ref)
        contents = load(conanfile_path)
        self.assertIn("#Comment for rev2", contents)

    def test_update_from_same_revision_test(self):
        """ If I specify conan install ref#revision --update it has to update to the latest binary
        of the same recipe.
        """
        conanfile = """
import time
import os
from conans import ConanFile, tools
class ConanFileToolsTest(ConanFile):
        
        def package(self):
            tools.save(os.path.join(self.package_folder, "file.txt"), str(time.time()))
"""
        ref = ConanFileReference.loads("Hello/0.1@lasote/stable")

        # Generate recipe rev 1 + package rev1 to remote0
        self.client.save({"conanfile.py": conanfile})
        self.client.run("create . %s" % str(ref))
        self.client.run("upload %s -r=remote0 --all" % str(ref))
        rev1 = self.client.get_revision(ref)
        package_ref = PackageReference(ref.copy_with_rev(rev1), NO_SETTINGS_PACKAGE_ID)
        prev1 = self.client.servers["remote0"].paths.get_last_package_revision(package_ref).revision

        # Use another client to install the only binary revision for ref
        client2 = TestClient(servers=self.servers, users=self.users)
        client2.run("install %s" % str(ref))

        # Generate recipe rev1 + package rev2 to remote0
        time.sleep(1)
        self.client.run("remove %s -f" % str(ref))
        self.client.run("create . %s" % str(ref))
        self.client.run("upload %s -r=remote0 --all" % str(ref))
        rev1_ = self.client.get_revision(ref)
        self.assertEquals(rev1, rev1_)
        package_ref = PackageReference(ref.copy_with_rev(rev1), NO_SETTINGS_PACKAGE_ID)
        prev2 = self.client.servers["remote0"].paths.get_last_package_revision(package_ref).revision
        self.assertNotEqual(prev1, prev2)  # Verify a new package revision is uploaded

        # Generate another recipe revision (and also bin revision)
        self.client.save({"conanfile.py": conanfile + "\n"})
        self.client.run("create . %s" % str(ref))
        self.client.run("upload %s -r=remote0 --all" % str(ref))

        # So, from client2 I install --update pinning the first recipe revision,
        # I don't want the recipe to be updated but the binary
        client2.run("install %s#%s --update" % (str(ref), rev1))
        self.assertNotIn("Hello/0.1@lasote/stable from 'remote0' - Updated", client2.out)
        self.assertIn("Hello/0.1@lasote/stable:"
                      "5ab84d6acfe1f23c4fae0ab88f26e3a396351ac9 - Update", client2.out)
        prev_ = client2.get_package_revision(package_ref)
        self.assertEquals(prev_, prev2)

    def old_server_new_client_simple_test(self):
        """Use this test as an example to try some new behavior with revisions against a server
        without revisions"""
        old_server = TestServer(server_capabilities=[])
        users = {"old_server": [("lasote", "mypass")]}
        client = TestClient(servers={"old_server": old_server}, users=users)
        conanfile = '''
from conans import ConanFile

class HelloConan(ConanFile):
    def build(self):
        self.output.warn("Revision 1")        
'''
        ref = ConanFileReference.loads("lib/1.0@lasote/testing")
        client.save({"conanfile.py": conanfile + "\n"})
        client.run("create . %s" % str(ref))
        client.run("upload %s -r=old_server --all" % str(ref))
        client.run("remove %s -f" % str(ref))
        client.run("install %s" % str(ref))
        package_ref = PackageReference(ref.copy_with_rev(DEFAULT_REVISION_V1),
                                       NO_SETTINGS_PACKAGE_ID)
        prev2 = client.servers["old_server"].paths.get_last_package_revision(package_ref)
        self.assertEquals(prev2.revision, DEFAULT_REVISION_V1)



import unittest
from conans.test.utils.tools import TestClient
from conans.paths import CONANFILE
from conans.test.utils.conanfile import TestConanFile


class VersionRangesErrorTest(unittest.TestCase):
    def verbose_version_test(self):
        client = TestClient()
        conanfile = TestConanFile("MyPkg", "0.1", requires=["MyOtherPkg/[~0.1]@user/testing"])
        client.save({CONANFILE: str(conanfile)})
        error = client.run("install . --build", ignore_error=True)
        self.assertTrue(error)
        self.assertIn("from requirement 'MyOtherPkg/[~0.1]@user/testing'", client.user_io.out)

    def werror_fail_test(self):
        client = TestClient()

        def add(name, version, requires=None):
            conanfile = TestConanFile(name, version, requires=requires)
            client.save({CONANFILE: str(conanfile)})
            client.run("export . user/testing")

        add("MyPkg1", "0.1.0")
        add("MyPkg1", "0.2.0")
        add("MyPkg2", "0.1", ["MyPkg1/[~0.1]@user/testing"])
        add("MyPkg3", "0.1", ["MyPkg1/[~0.2]@user/testing", "MyPkg2/[~0.1]@user/testing"])

        error = client.run("install . --build", ignore_error=True)
        self.assertTrue(error)
        self.assertNotIn("WARN: Version range '~0.1' required", client.out)
        self.assertIn("ERROR: Version range '~0.1' required by 'MyPkg2/0.1@user/testing' "
                      "not valid for downstream requirement 'MyPkg1/0.2.0@user/testing'",
                      client.out)



import os
import platform
import unittest

import mock
from nose.plugins.attrib import attr
from parameterized import parameterized

from conans.client.build.msbuild import MSBuild
from conans.errors import ConanException
from conans.model.version import Version
from conans.paths import CONANFILE
from conans.test.utils.conanfile import MockSettings, MockConanfile
from conans.test.utils.tools import TestClient
from conans.test.utils.visual_project_files import get_vs_project_files
from conans.client import tools


class MSBuildTest(unittest.TestCase):

    def dont_mess_with_build_type_test(self):
        settings = MockSettings({"build_type": "Debug",
                                 "compiler": "Visual Studio",
                                 "arch": "x86_64",
                                 "compiler.runtime": "MDd"})
        conanfile = MockConanfile(settings)
        msbuild = MSBuild(conanfile)
        self.assertEquals(msbuild.build_env.flags, [])
        template = msbuild._get_props_file_contents()

        self.assertNotIn("-Ob0", template)
        self.assertNotIn("-Od", template)

        msbuild.build_env.flags = ["-Zi"]
        template = msbuild._get_props_file_contents()

        self.assertNotIn("-Ob0", template)
        self.assertNotIn("-Od", template)
        self.assertIn("-Zi", template)
        self.assertIn("<RuntimeLibrary>MultiThreadedDebugDLL</RuntimeLibrary>", template)

    def without_runtime_test(self):
        settings = MockSettings({"build_type": "Debug",
                                 "compiler": "Visual Studio",
                                 "arch": "x86_64"})
        conanfile = MockConanfile(settings)
        msbuild = MSBuild(conanfile)
        template = msbuild._get_props_file_contents()
        self.assertNotIn("<RuntimeLibrary>", template)

    @attr('slow')
    @unittest.skipUnless(platform.system() == "Windows", "Requires MSBuild")
    def build_vs_project_test(self):
        conan_build_vs = """
from conans import ConanFile, MSBuild

class HelloConan(ConanFile):
    name = "Hello"
    version = "1.2.1"
    exports = "*"
    settings = "os", "build_type", "arch", "compiler", "cppstd"

    def build(self):
        msbuild = MSBuild(self)
        msbuild.build("MyProject.sln")

    def package(self):
        self.copy(pattern="*.exe")

"""
        client = TestClient()

        # Test cpp standard stuff

        files = get_vs_project_files(std="cpp17_2015")
        files[CONANFILE] = conan_build_vs

        client.save(files)
        error = client.run('create . Hello/1.2.1@lasote/stable -s cppstd=11 -s '
                           'compiler="Visual Studio" -s compiler.version=14', ignore_error=True)
        self.assertTrue(error)
        client.run('create . Hello/1.2.1@lasote/stable -s cppstd=17 '
                   '-s compiler="Visual Studio" -s compiler.version=14')
        self.assertIn("Copied 1 '.exe' file: MyProject.exe", client.user_io.out)

        files = get_vs_project_files()
        files[CONANFILE] = conan_build_vs

        # Try to not update the project
        client.client_cache._conan_config = None  # Invalidate cached config
        tools.replace_in_file(client.client_cache.conan_conf_path, "[general]",
                              "[general]\nskip_vs_projects_upgrade = True")
        client.save(files, clean_first=True)
        client.run("create . Hello/1.2.1@lasote/stable --build")
        self.assertNotIn("devenv", client.user_io.out)
        self.assertIn("Skipped sln project upgrade", client.user_io.out)

        # Try with x86_64
        client.save(files)
        client.run("export . lasote/stable")
        client.run("install Hello/1.2.1@lasote/stable --build -s arch=x86_64")
        self.assertIn("Release|x64", client.user_io.out)
        self.assertIn("Copied 1 '.exe' file: MyProject.exe", client.user_io.out)

        # Try with x86
        client.save(files, clean_first=True)
        client.run("export . lasote/stable")
        client.run("install Hello/1.2.1@lasote/stable --build -s arch=x86")
        self.assertIn("Release|x86", client.user_io.out)
        self.assertIn("Copied 1 '.exe' file: MyProject.exe", client.user_io.out)

        # Try with x86 debug
        client.save(files, clean_first=True)
        client.run("export . lasote/stable")
        client.run("install Hello/1.2.1@lasote/stable --build -s arch=x86 -s build_type=Debug")
        self.assertIn("Debug|x86", client.user_io.out)
        self.assertIn("Copied 1 '.exe' file: MyProject.exe", client.user_io.out)

    def custom_properties_test(self):
        settings = MockSettings({"build_type": "Debug",
                                 "compiler": "Visual Studio",
                                 "arch": "x86_64"})
        conanfile = MockConanfile(settings)
        msbuild = MSBuild(conanfile)
        command = msbuild.get_command("project_file.sln", properties={"MyProp1": "MyValue1",
                                                                      "MyProp2": "MyValue2"})
        self.assertIn('/p:MyProp1="MyValue1"', command)
        self.assertIn('/p:MyProp2="MyValue2"', command)

    @unittest.skipUnless(platform.system() == "Windows", "Requires MSBuild")
    def reuse_msbuild_object_test(self):
        # https://github.com/conan-io/conan/issues/2865
        conan_build_vs = """
from conans import ConanFile, MSBuild

class HelloConan(ConanFile):
    name = "Hello"
    version = "1.2.1"
    exports = "*"
    settings = "os", "build_type", "arch", "compiler", "cppstd"

    def configure(self):
        del self.settings.compiler.runtime
        del self.settings.build_type

    def build(self):
        msbuild = MSBuild(self)
        msbuild.build("MyProject.sln", build_type="Release")
        msbuild.build("MyProject.sln", build_type="Debug")
        self.output.info("build() completed")
"""
        client = TestClient()
        files = get_vs_project_files()
        files[CONANFILE] = conan_build_vs

        client.save(files)
        client.run("create . danimtb/testing")
        self.assertIn("build() completed", client.out)

    @unittest.skipUnless(platform.system() == "Windows", "Requires MSBuild")
    def binary_logging_on_test(self):
        settings = MockSettings({"build_type": "Debug",
                                 "compiler": "Visual Studio",
                                 "compiler.version": "15",
                                 "arch": "x86_64",
                                 "compiler.runtime": "MDd"})
        conanfile = MockConanfile(settings)
        msbuild = MSBuild(conanfile)
        command = msbuild.get_command("dummy.sln", output_binary_log=True)
        self.assertIn("/bl", command)

    @unittest.skipUnless(platform.system() == "Windows", "Requires MSBuild")
    def binary_logging_on_with_filename_test(self):
        bl_filename = "a_special_log.log"
        settings = MockSettings({"build_type": "Debug",
                                 "compiler": "Visual Studio",
                                 "compiler.version": "15",
                                 "arch": "x86_64",
                                 "compiler.runtime": "MDd"})
        conanfile = MockConanfile(settings)
        msbuild = MSBuild(conanfile)
        command = msbuild.get_command("dummy.sln", output_binary_log=bl_filename)
        expected_command = '/bl:"%s"' % bl_filename
        self.assertIn(expected_command, command)

    def binary_logging_off_explicit_test(self):
        settings = MockSettings({"build_type": "Debug",
                                 "compiler": "Visual Studio",
                                 "compiler.version": "15",
                                 "arch": "x86_64",
                                 "compiler.runtime": "MDd"})
        conanfile = MockConanfile(settings)
        msbuild = MSBuild(conanfile)
        command = msbuild.get_command("dummy.sln", output_binary_log=False)
        self.assertNotIn("/bl", command)

    def binary_logging_off_implicit_test(self):
        settings = MockSettings({"build_type": "Debug",
                                 "compiler": "Visual Studio",
                                 "compiler.version": "15",
                                 "arch": "x86_64",
                                 "compiler.runtime": "MDd"})
        conanfile = MockConanfile(settings)
        msbuild = MSBuild(conanfile)
        command = msbuild.get_command("dummy.sln")
        self.assertNotIn("/bl", command)

    @unittest.skipUnless(platform.system() == "Windows", "Requires MSBuild")
    @mock.patch("conans.client.build.msbuild.MSBuild.get_version")
    def binary_logging_not_supported_test(self, mock_get_version):
        mock_get_version.return_value = Version("14")

        mocked_settings = MockSettings({"build_type": "Debug",
                                        "compiler": "Visual Studio",
                                        "compiler.version": "15",
                                        "arch": "x86_64",
                                        "compiler.runtime": "MDd"})
        conanfile = MockConanfile(mocked_settings)
        except_text = "MSBuild version detected (14) does not support 'output_binary_log' ('/bl')"
        msbuild = MSBuild(conanfile)

        with self.assertRaises(ConanException) as exc:
            msbuild.get_command("dummy.sln", output_binary_log=True)
        self.assertIn(except_text, str(exc.exception))

    @unittest.skipUnless(platform.system() == "Windows", "Requires MSBuild")
    def get_version_test(self):
        settings = MockSettings({"build_type": "Debug",
                                 "compiler": "Visual Studio",
                                 "compiler.version": "15",
                                 "arch": "x86_64",
                                 "compiler.runtime": "MDd"})
        version = MSBuild.get_version(settings)
        self.assertRegexpMatches(version, "(\d+\.){2,3}\d+")
        self.assertGreater(version, "15.1")

    @parameterized.expand([("True",), ("'my_log.binlog'",)])
    @unittest.skipUnless(platform.system() == "Windows", "Requires MSBuild")
    def binary_log_build_test(self, value):
        conan_build_vs = """
from conans import ConanFile, MSBuild

class HelloConan(ConanFile):
    name = "Hello"
    version = "1.2.1"
    exports = "*"
    settings = "os", "build_type", "arch", "compiler"

    def build(self):
        msbuild = MSBuild(self)
        msbuild.build("MyProject.sln", output_binary_log=%s)
"""
        client = TestClient()
        files = get_vs_project_files()
        files[CONANFILE] = conan_build_vs % value
        client.save(files)
        client.run("install . -s compiler=\"Visual Studio\" -s compiler.version=15")
        client.run("build .")

        if value == "'my_log.binlog'":
            log_name = value[1:1]
            flag = "/bl:%s" % log_name
        else:
            log_name = "msbuild.binlog"
            flag = "/bl"

        self.assertIn(flag, client.out)
        log_path = os.path.join(client.current_folder, log_name)
        self.assertTrue(os.path.exists(log_path))



from conans.model.version import Version


def cppstd_flag(compiler, compiler_version, cppstd):
    if not compiler or not compiler_version or not cppstd:
        return ""
    func = {"gcc": _cppstd_gcc,
            "clang": _cppstd_clang,
            "apple-clang": _cppstd_apple_clang,
            "Visual Studio": _cppstd_visualstudio}.get(str(compiler), None)
    flag = None
    if func:
        flag = func(str(compiler_version), str(cppstd))
    return flag


def cppstd_default(compiler, compiler_version):
    default = {"gcc": _gcc_cppstd_default(compiler_version),
               "clang": _clang_cppstd_default(compiler_version),
               "apple-clang": "gnu98",  # Confirmed in apple-clang 9.1 with a simple "auto i=1;"
               "Visual Studio": _visual_cppstd_default(compiler_version)}.get(str(compiler), None)
    return default


def _clang_cppstd_default(compiler_version):
    # Official docs are wrong, in 6.0 the default is gnu14 to follow gcc's choice
    return "gnu98" if Version(compiler_version) < "6" else "gnu14"


def _gcc_cppstd_default(compiler_version):
    return "gnu98" if Version(compiler_version) < "6" else "gnu14"


def _visual_cppstd_default(compiler_version):
    if Version(compiler_version) >= "14":  # VS 2015 update 3 only
        return "14"
    return None


def _cppstd_visualstudio(visual_version, cppstd):
    # https://docs.microsoft.com/en-us/cpp/build/reference/std-specify-language-standard-version
    v14 = None
    v17 = None
    v20 = None

    if Version(visual_version) >= "14":
        v14 = "c++14"
        v17 = "c++latest"
    if Version(visual_version) >= "15":
        v17 = "c++17"
        v20 = "c++latest"

    flag = {"14": v14, "17": v17, "20": v20}.get(str(cppstd), None)
    return "/std:%s" % flag if flag else None


def _cppstd_apple_clang(clang_version, cppstd):
    """
    Inspired in:
    https://github.com/Kitware/CMake/blob/master/Modules/Compiler/AppleClang-CXX.cmake
    """

    v98 = vgnu98 = v11 = vgnu11 = v14 = vgnu14 = v17 = vgnu17 = v20 = vgnu20 = None

    if Version(clang_version) >= "4.0":
        v98 = "c++98"
        vgnu98 = "gnu++98"
        v11 = "c++11"
        vgnu11 = "gnu++11"

    if Version(clang_version) >= "6.1":
        v14 = "c++14"
        vgnu14 = "gnu++14"
    elif Version(clang_version) >= "5.1":
        v14 = "c++1y"
        vgnu14 = "gnu++1y"

    if Version(clang_version) >= "6.1":
        v17 = "c++1z"
        vgnu17 = "gnu++1z"

    if Version(clang_version) >= "9.1":
        # Not confirmed that it didn't work before 9.1 but 1z is still valid, so we are ok
        v17 = "c++17"
        vgnu17 = "gnu++17"

    flag = {"98": v98, "gnu98": vgnu98,
            "11": v11, "gnu11": vgnu11,
            "14": v14, "gnu14": vgnu14,
            "17": v17, "gnu17": vgnu17,
            "20": v20, "gnu20": vgnu20}.get(cppstd, None)

    return "-std=%s" % flag if flag else None


def _cppstd_clang(clang_version, cppstd):
    """
    Inspired in:
    https://github.com/Kitware/CMake/blob/
    1fe2dc5ef2a1f262b125a2ba6a85f624ce150dd2/Modules/Compiler/Clang-CXX.cmake

    https://clang.llvm.org/cxx_status.html
    """
    v98 = vgnu98 = v11 = vgnu11 = v14 = vgnu14 = v17 = vgnu17 = v20 = vgnu20 = None

    if Version(clang_version) >= "2.1":
        v98 = "c++98"
        vgnu98 = "gnu++98"

    if Version(clang_version) >= "3.1":
        v11 = "c++11"
        vgnu11 = "gnu++11"
    elif Version(clang_version) >= "2.1":
        v11 = "c++0x"
        vgnu11 = "gnu++0x"

    if Version(clang_version) >= "3.5":
        v14 = "c++14"
        vgnu14 = "gnu++14"
    elif Version(clang_version) >= "3.4":
        v14 = "c++1y"
        vgnu14 = "gnu++1y"

    if Version(clang_version) >= "5":
        v17 = "c++17"
        vgnu17 = "gnu++17"
    elif Version(clang_version) >= "3.5":
        v17 = "c++1z"
        vgnu17 = "gnu++1z"

    if Version(clang_version) >= "6":
        v20 = "c++2a"
        vgnu20 = "gnu++2a"

    flag = {"98": v98, "gnu98": vgnu98,
            "11": v11, "gnu11": vgnu11,
            "14": v14, "gnu14": vgnu14,
            "17": v17, "gnu17": vgnu17,
            "20": v20, "gnu20": vgnu20}.get(cppstd, None)
    return "-std=%s" % flag if flag else None


def _cppstd_gcc(gcc_version, cppstd):
    """https://github.com/Kitware/CMake/blob/master/Modules/Compiler/GNU-CXX.cmake"""
    # https://gcc.gnu.org/projects/cxx-status.html
    v98 = vgnu98 = v11 = vgnu11 = v14 = vgnu14 = v17 = vgnu17 = v20 = vgnu20 = None

    if Version(gcc_version) >= "3.4":
        v98 = "c++98"
        vgnu98 = "gnu++98"

    if Version(gcc_version) >= "4.7":
        v11 = "c++11"
        vgnu11 = "gnu++11"
    elif Version(gcc_version) >= "4.3":
        v11 = "c++0x"
        vgnu11 = "gnu++0x"

    if Version(gcc_version) >= "4.9":
        v14 = "c++14"
        vgnu14 = "gnu++14"
    elif Version(gcc_version) >= "4.8":
        v14 = "c++1y"
        vgnu14 = "gnu++1y"

    if Version(gcc_version) >= "5.1":
        v17 = "c++1z"
        vgnu17 = "gnu++1z"

    if Version(gcc_version) >= "5.2":  # Not sure if even in 5.1 gnu17 is valid, but gnu1z is
        v17 = "c++17"
        vgnu17 = "gnu++17"

    if Version(gcc_version) >= "8":
        v20 = "c++2a"
        vgnu20 = "gnu++2a"

    flag = {"98": v98, "gnu98": vgnu98,
            "11": v11, "gnu11": vgnu11,
            "14": v14, "gnu14": vgnu14,
            "17": v17, "gnu17": vgnu17,
            "20": v20, "gnu20": vgnu20}.get(cppstd)
    return "-std=%s" % flag if flag else None



import os
import shutil
import unittest
from collections import namedtuple

from conans.client.tools import chdir
from conans.model.ref import ConanFileReference
from conans.test.utils.tools import TestServer, TestClient
from conans.util.files import save, load
from conans.client.cmd.export import _replace_scm_data_in_conanfile
from conans.client.loader import _parse_conanfile
from conans.test.utils.test_files import temp_folder
from conans.model.scm import SCMData


class ReplaceSCMDataInConanfileTest(unittest.TestCase):
    conanfile = """
from conans import ConanFile

class ConanLib(ConanFile):
    name = "lib"
    version = "0.1"
    scm = {{"revision": "{revision}",
           "type": "git",
           "url": "{url}"}}

    {after_scm}

{after_recipe}
"""

    def run(self, *args, **kwargs):
        self.tmp_folder = temp_folder()
        self.conanfile_path = os.path.join(self.tmp_folder, 'conanfile.py')
        try:
            super(ReplaceSCMDataInConanfileTest, self).run(*args, **kwargs)
        finally:
            shutil.rmtree(self.tmp_folder)

    def _do_actual_test(self, scm_data, after_scm, after_recipe):
        target_conanfile = self.conanfile.format(url=scm_data['url'],
                                                 revision=scm_data['revision'],
                                                 after_scm=after_scm,
                                                 after_recipe=after_recipe)
        save(self.conanfile_path, content=self.conanfile.format(url='auto', revision='auto',
                                                                after_scm=after_scm,
                                                                after_recipe=after_recipe))
        scm_data = SCMData(conanfile=namedtuple('_', 'scm')(scm=scm_data))
        _replace_scm_data_in_conanfile(self.conanfile_path, scm_data)
        self.assertEqual(load(self.conanfile_path), target_conanfile)
        _parse_conanfile(conan_file_path=self.conanfile_path)  # Check that the resulting file is valid python code.

    def test_conanfile_after_scm(self):
        scm_data = {'type': 'git',
                    'url': 'url_value',
                    'revision': 'revision_value'}
        after_scm = 'attrib = 23'
        after_recipe = ''
        self._do_actual_test(scm_data=scm_data, after_scm=after_scm, after_recipe=after_recipe)

    def test_conanfile_after_scm_and_recipe(self):
        scm_data = {'type': 'git',
                    'url': 'url_value',
                    'revision': 'revision_value'}
        after_scm = 'attrib = 23'
        after_recipe = 'another = 23'
        self._do_actual_test(scm_data=scm_data, after_scm=after_scm, after_recipe=after_recipe)

    def test_conanfile_after_recipe(self):
        scm_data = {'type': 'git',
                    'url': 'url_value',
                    'revision': 'revision_value'}
        after_scm = ''
        after_recipe = 'another = 23'
        self._do_actual_test(scm_data=scm_data, after_scm=after_scm, after_recipe=after_recipe)

    def test_conanfile_none(self):
        scm_data = {'type': 'git',
                    'url': 'url_value',
                    'revision': 'revision_value'}
        after_scm = ''
        after_recipe = ''
        self._do_actual_test(scm_data=scm_data, after_scm=after_scm, after_recipe=after_recipe)


class SCMUpload(unittest.TestCase):

    def scm_sources_test(self):
        """ Test conan_sources.tgz is deleted in server when removing 'exports_sources' and using
        'scm'"""
        conanfile = """from conans import ConanFile
class TestConan(ConanFile):
    name = "test"
    version = "1.0"
"""
        exports_sources = """
    exports_sources = "include/*"
"""
        servers = {"upload_repo": TestServer([("*/*@*/*", "*")], [("*/*@*/*", "*")],
                                             users={"lasote": "mypass"})}
        client = TestClient(servers=servers, users={"upload_repo": [("lasote", "mypass")]})
        client.save({"conanfile.py": conanfile + exports_sources, "include/file": "content"})
        client.run("create . danimtb/testing")
        client.run("upload test/1.0@danimtb/testing -r upload_repo")
        self.assertIn("Uploading conan_sources.tgz", client.out)
        conan_ref = ConanFileReference("test", "1.0", "danimtb", "testing")
        export_sources_path = os.path.join(servers["upload_repo"].paths.export(conan_ref),
                                           "conan_sources.tgz")
        self.assertTrue(os.path.exists(export_sources_path))

        scm = """
    scm = {"type": "git",
           "url": "auto",
           "revision": "auto"}
"""
        client.save({"conanfile.py": conanfile + scm})
        with chdir(client.current_folder):
            client.runner("git init")
            client.runner('git config user.email "you@example.com"')
            client.runner('git config user.name "Your Name"')
            client.runner("git remote add origin https://github.com/fake/fake.git")
            client.runner("git add .")
            client.runner("git commit -m \"initial commit\"")
        client.run("create . danimtb/testing")
        self.assertIn("Repo origin deduced by 'auto': https://github.com/fake/fake.git", client.out)
        client.run("upload test/1.0@danimtb/testing -r upload_repo")
        self.assertNotIn("Uploading conan_sources.tgz", client.out)
        export_sources_path = os.path.join(servers["upload_repo"].paths.export(conan_ref),
                                           "conan_sources.tgz")
        self.assertFalse(os.path.exists(export_sources_path))



import unittest

from conans.client.recorder.search_recorder import SearchRecorder
from conans.model.ref import ConanFileReference

_ref = ConanFileReference.loads


class SearchRecorderTest(unittest.TestCase):

    def setUp(self):
        self.recorder = SearchRecorder()

    def empty_test(self):
        info = self.recorder.get_info()
        expected_result = {'error': False, 'results': []}
        self.assertEqual(expected_result, info)

    def sequential_test(self):
        self.recorder.add_recipe("remote1", _ref("fake/0.1@user/channel"))
        self.recorder.add_package("remote1", _ref("fake/0.1@user/channel"), "fake_package_id",
                                  "fake_options", "fake_settings", "fake_requires", False)
        self.recorder.add_recipe("remote2", _ref("fakefake/0.1@user/channel"))
        self.recorder.add_package("remote2", _ref("fakefake/0.1@user/channel"), "fakefake_package_id1",
                                  "fakefake_options1", "fakefake_settings1", "fakefake_requires1",
                                  False)
        self.recorder.add_package("remote2", _ref("fakefake/0.1@user/channel"), "fakefake_package_id2",
                                  "fakefake_options2", "fakefake_settings2", "fakefake_requires2",
                                  False)
        info = self.recorder.get_info()
        expected_result = {
                               "error": False,
                               "results": [
                                   {
                                        "remote": "remote1",
                                        "items": [
                                            {
                                                "recipe": {
                                                    "id": "fake/0.1@user/channel"
                                                },
                                                "packages": [
                                                    {
                                                        "id": "fake_package_id",
                                                        "options": "fake_options",
                                                        "settings": "fake_settings",
                                                        "requires": "fake_requires",
                                                        "outdated": False
                                                    }
                                                ]
                                            }
                                        ]
                                   },
                                   {
                                       "remote": "remote2",
                                       "items": [
                                           {
                                               "recipe": {
                                                   "id": "fakefake/0.1@user/channel"
                                               },
                                               "packages": [
                                                   {
                                                       "id": "fakefake_package_id1",
                                                       "options": "fakefake_options1",
                                                       "settings": "fakefake_settings1",
                                                       "requires": "fakefake_requires1",
                                                       "outdated": False
                                                   },
                                                   {
                                                       "id": "fakefake_package_id2",
                                                       "options": "fakefake_options2",
                                                       "settings": "fakefake_settings2",
                                                       "requires": "fakefake_requires2",
                                                       "outdated": False
                                                   }
                                               ]
                                           }
                                       ]
                                   }
                               ]
        }
        self.assertEqual(expected_result, info)

    def unordered_test(self):
        self.recorder.add_recipe("my_remote1", _ref("fake1/0.1@user/channel"))
        self.recorder.add_recipe("my_remote2", _ref("fake2/0.1@user/channel"))
        self.recorder.add_recipe("my_remote3", _ref("fake3/0.1@user/channel"))
        self.recorder.add_package("my_remote1", _ref("fake1/0.1@user/channel"), "fake1_package_id1",
                                  "fake1_options1", "fake1_settings1", "fake1_requires1", False)
        self.recorder.add_package("my_remote2", _ref("fake2/0.1@user/channel"), "fake2_package_id1",
                                  "fake2_options1", "fake2_settings1", "fake2_requires1", False)
        self.recorder.add_package("my_remote2", _ref("fake2/0.1@user/channel"), "fake2_package_id2",
                                  "fake2_options2", "fake2_settings2", "fake2_requires2", False)
        info = self.recorder.get_info()
        expected_result = {
                            "error": False,
                            "results": [
                                {
                                    "remote": "my_remote1",
                                    "items": [
                                        {
                                            "recipe": {
                                                "id": "fake1/0.1@user/channel"
                                            },
                                            "packages": [
                                                {
                                                    "id": "fake1_package_id1",
                                                    "options": "fake1_options1",
                                                    "settings": "fake1_settings1",
                                                    "requires": "fake1_requires1",
                                                    "outdated": False
                                                }
                                            ]
                                        }
                                    ]
                                },
                                {
                                    "remote": "my_remote2",
                                    "items": [
                                        {
                                            "recipe": {
                                                "id": "fake2/0.1@user/channel"
                                            },
                                            "packages": [
                                                {
                                                    "id": "fake2_package_id1",
                                                    "options": "fake2_options1",
                                                    "settings": "fake2_settings1",
                                                    "requires": "fake2_requires1",
                                                    "outdated": False
                                                },
                                                {
                                                    "id": "fake2_package_id2",
                                                    "options": "fake2_options2",
                                                    "settings": "fake2_settings2",
                                                    "requires": "fake2_requires2",
                                                    "outdated": False
                                                }
                                            ]
                                        }
                                    ]
                                },
                                {
                                    "remote": "my_remote3",
                                    "items": [
                                        {
                                            "recipe": {
                                                "id": "fake3/0.1@user/channel"
                                            },
                                            "packages": []
                                        }
                                    ]
                                }
                            ]
                        }
        self.assertEqual(expected_result, info)

    def without_packages_test(self):
        self.recorder.add_recipe("my_remote1", _ref("fake1/0.1@user/channel"), None)
        self.recorder.add_recipe("my_remote2", _ref("fake2/0.1@user/channel"), None)
        self.recorder.add_recipe("my_remote3", _ref("fake3/0.1@user/channel"), None)
        info = self.recorder.get_info()
        expected_result = {
                            "error": False,
                            "results": [
                                {
                                    "remote": "my_remote1",
                                    "items": [
                                        {
                                            "recipe": {
                                                "id": "fake1/0.1@user/channel"
                                            }
                                        }
                                    ]
                                },
                                {
                                    "remote": "my_remote2",
                                    "items": [
                                        {
                                            "recipe": {
                                                "id": "fake2/0.1@user/channel"
                                            }
                                        }
                                    ]
                                },
                                {
                                    "remote": "my_remote3",
                                    "items": [
                                        {
                                            "recipe": {
                                                "id": "fake3/0.1@user/channel"
                                            }
                                        }
                                    ]
                                }
                            ]
                        }
        self.assertEqual(expected_result, info)



import copy
import re
import subprocess

from conans.client.build.visual_environment import (VisualStudioBuildEnvironment,
                                                    vs_build_type_flags, vs_std_cpp)
from conans.client.tools.oss import cpu_count
from conans.client.tools.win import vcvars_command
from conans.errors import ConanException
from conans.model.version import Version
from conans.util.env_reader import get_env
from conans.util.files import tmp_file, decode_text
from conans.model.conan_file import ConanFile
from conans.client import tools


class MSBuild(object):

    def __init__(self, conanfile):
        if isinstance(conanfile, ConanFile):
            self._conanfile = conanfile
            self._settings = self._conanfile.settings
            self._output = self._conanfile.output
            self.build_env = VisualStudioBuildEnvironment(self._conanfile,
                                                          with_build_type_flags=False)
        else:  # backwards compatible with build_sln_command
            self._settings = conanfile
            self.build_env = None

    def build(self, project_file, targets=None, upgrade_project=True, build_type=None, arch=None,
              parallel=True, force_vcvars=False, toolset=None, platforms=None, use_env=True,
              vcvars_ver=None, winsdk_version=None, properties=None, output_binary_log=None):

        self.build_env.parallel = parallel

        with tools.environment_append(self.build_env.vars):
            # Path for custom properties file
            props_file_contents = self._get_props_file_contents()
            with tmp_file(props_file_contents) as props_file_path:
                vcvars = vcvars_command(self._conanfile.settings, force=force_vcvars,
                                        vcvars_ver=vcvars_ver, winsdk_version=winsdk_version)
                command = self.get_command(project_file, props_file_path,
                                           targets=targets, upgrade_project=upgrade_project,
                                           build_type=build_type, arch=arch, parallel=parallel,
                                           toolset=toolset, platforms=platforms,
                                           use_env=use_env, properties=properties, output_binary_log=output_binary_log)
                command = "%s && %s" % (vcvars, command)
                return self._conanfile.run(command)

    def get_command(self, project_file, props_file_path=None, targets=None, upgrade_project=True,
                    build_type=None, arch=None, parallel=True, toolset=None, platforms=None,
                    use_env=False, properties=None, output_binary_log=None):

        targets = targets or []
        properties = properties or {}
        command = []

        if upgrade_project and not get_env("CONAN_SKIP_VS_PROJECTS_UPGRADE", False):
            command.append('devenv "%s" /upgrade &&' % project_file)
        else:
            self._output.info("Skipped sln project upgrade")

        build_type = build_type or self._settings.get_safe("build_type")
        arch = arch or self._settings.get_safe("arch")
        if not build_type:
            raise ConanException("Cannot build_sln_command, build_type not defined")
        if not arch:
            raise ConanException("Cannot build_sln_command, arch not defined")

        command.append('msbuild "%s" /p:Configuration="%s"' % (project_file, build_type))
        msvc_arch = {'x86': 'x86',
                     'x86_64': 'x64',
                     'armv7': 'ARM',
                     'armv8': 'ARM64'}
        if platforms:
            msvc_arch.update(platforms)
        msvc_arch = msvc_arch.get(str(arch))
        try:
            sln = tools.load(project_file)
            pattern = re.compile(r"GlobalSection\(SolutionConfigurationPlatforms\)"
                                 r"(.*?)EndGlobalSection", re.DOTALL)
            solution_global = pattern.search(sln).group(1)
            lines = solution_global.splitlines()
            lines = [s.split("=")[0].strip() for s in lines]
        except Exception:
            pass  # TODO: !!! what are we catching here? tools.load? .group(1)? .splitlines?
        else:
            config = "%s|%s" % (build_type, msvc_arch)
            if config not in "".join(lines):
                self._output.warn("***** The configuration %s does not exist in this solution *****" % config)
                self._output.warn("Use 'platforms' argument to define your architectures")

        if output_binary_log:
            msbuild_version = MSBuild.get_version(self._settings)
            if msbuild_version >= "15.3":  # http://msbuildlog.com/
                command.append('/bl' if isinstance(output_binary_log, bool)
                               else '/bl:"%s"' % output_binary_log)
            else:
                raise ConanException("MSBuild version detected (%s) does not support "
                                     "'output_binary_log' ('/bl')" % msbuild_version)

        if use_env:
            command.append('/p:UseEnv=true')

        if msvc_arch:
            command.append('/p:Platform="%s"' % msvc_arch)

        if parallel:
            command.append('/m:%s' % cpu_count(output=self._output))

        if targets:
            command.append("/target:%s" % ";".join(targets))

        if toolset:
            command.append('/p:PlatformToolset="%s"' % toolset)

        if props_file_path:
            command.append('/p:ForceImportBeforeCppTargets="%s"' % props_file_path)

        for name, value in properties.items():
            command.append('/p:%s="%s"' % (name, value))

        return " ".join(command)

    def _get_props_file_contents(self):
        # how to specify runtime in command line:
        # https://stackoverflow.com/questions/38840332/msbuild-overrides-properties-while-building-vc-project
        runtime_library = {"MT": "MultiThreaded",
                           "MTd": "MultiThreadedDebug",
                           "MD": "MultiThreadedDLL",
                           "MDd": "MultiThreadedDebugDLL"}.get(self._settings.get_safe("compiler.runtime"), "")

        if self.build_env:
            # Take the flags from the build env, the user was able to alter them if needed
            flags = copy.copy(self.build_env.flags)
            flags.append(self.build_env.std)
        else:  # To be removed when build_sln_command is deprecated
            flags = vs_build_type_flags(self._settings, with_flags=False)
            flags.append(vs_std_cpp(self._settings))

        flags_str = " ".join(list(filter(None, flags))) # Removes empty and None elements
        additional_node = "<AdditionalOptions>" \
                          "{} %(AdditionalOptions)" \
                          "</AdditionalOptions>".format(flags_str) if flags_str else ""
        runtime_node = "<RuntimeLibrary>" \
                       "{}" \
                       "</RuntimeLibrary>".format(runtime_library) if runtime_library else ""
        template = """<?xml version="1.0" encoding="utf-8"?>
<Project xmlns="http://schemas.microsoft.com/developer/msbuild/2003">
  <ItemDefinitionGroup>
    <ClCompile>
      {runtime_node}
      {additional_node}
    </ClCompile>
  </ItemDefinitionGroup>
</Project>""".format(**{"runtime_node": runtime_node,
                        "additional_node": additional_node})
        return template

    @staticmethod
    def get_version(settings):
        msbuild_cmd = "msbuild -version"
        vcvars = vcvars_command(settings)
        command = "%s && %s" % (vcvars, msbuild_cmd)
        try:
            out, err = subprocess.Popen(command, stdout=subprocess.PIPE, shell=True).communicate()
            version_line = decode_text(out).split("\n")[-1]
            prog = re.compile("(\d+\.){2,3}\d+")
            result = prog.match(version_line).group()
            return Version(result)
        except Exception as e:
            raise ConanException("Error retrieving MSBuild version: '{}'".format(e))



import platform
import os
import yaml

from collections import OrderedDict

from conans.errors import ConanException
from conans.util.files import load, save, mkdir
from conans.model.ref import ConanFileReference


class LocalPackage(object):
    def __init__(self, base_folder, install_folder, data):
        self._base_folder = base_folder
        self._install_folder = install_folder
        self._conanfile_folder = data.get("folder")  # The folder with the conanfile
        self._source_folder = data.get("source")  # By default the conanfile_folder
        self._build_folder = data.get("build", "")
        # package_folder can be None, then it will directly use build_folder
        self._package_folder = data.get("package", "")
        self._cmakedir = data.get("cmakedir")
        includedirs = data.get("includedirs", [])  # To override include dirs, mainly for build_folder
        self._includedirs = [includedirs] if not isinstance(includedirs, list) else includedirs
        libdirs = data.get("libdirs", [])  # To override libdirs...
        self._libdirs = [libdirs] if not isinstance(libdirs, list) else libdirs
        self.conanfile = None

    @property
    def root_folder(self):
        return os.path.abspath(os.path.join(self._base_folder, self._conanfile_folder))

    @property
    def conanfile_path(self):
        return os.path.join(self.root_folder, "conanfile.py")

    @property
    def build_folder(self):
        folder = self._evaluate(self._build_folder)
        if self._install_folder:
            pkg_folder = os.path.join(self._install_folder, self._conanfile_folder, folder)
        else:
            pkg_folder = os.path.join(self.root_folder, folder)
        mkdir(pkg_folder)
        return pkg_folder

    @property
    def package_folder(self):
        folder = self._evaluate(self._package_folder)
        if self._install_folder:
            pkg_folder = os.path.join(self._install_folder, self._conanfile_folder, folder)
        else:
            pkg_folder = os.path.join(self.root_folder, folder)
        mkdir(pkg_folder)
        return pkg_folder

    @property
    def install_folder(self):
        return self._evaluate(self._install_folder)

    @property
    def includedirs(self):
        return [os.path.join(self._base_folder, self._conanfile_folder, v)
                for v in self._includedirs]

    @property
    def libdirs(self):
        return [self._evaluate(v) for v in self._libdirs]

    def _evaluate(self, value):
        settings = self.conanfile.settings
        v = value.format(build_type=settings.get_safe("build_type"),
                         arch=settings.get_safe("arch"),
                         os=platform.system())
        try:
            result = eval('%s' % v)
        except:
            result = v
        return result

    @property
    def local_cmakedir(self):
        if not self._cmakedir:
            return self._conanfile_folder
        return os.path.join(self._conanfile_folder, self._cmakedir).replace("\\", "/")


WORKSPACE_FILE = "conanws.yml"


class Workspace(object):
    @staticmethod
    def get_workspace(folder, install_folder):
        if isinstance(folder, ConanFileReference):
            return None
        if not os.path.exists(folder):
            return None
        path = os.path.join(folder, WORKSPACE_FILE)
        if os.path.exists(path):
            return Workspace(path, install_folder)
        parent = os.path.dirname(folder)
        if parent and parent != folder:
            return Workspace.get_workspace(parent, install_folder)
        return None

    def generate(self):
        if self._generator == "cmake":
            template = """# conanws
cmake_minimum_required(VERSION 3.3)
project({name} CXX)

"""
            cmake = template.format(name=self._name)
            for _, workspace_package in self._workspace_packages.items():
                build_folder = workspace_package.build_folder
                build_folder = build_folder.replace("\\", "/")
                cmake += 'add_subdirectory(%s "%s")\n' % (workspace_package.local_cmakedir, build_folder)
            cmake_path = os.path.join(self._base_folder, "CMakeLists.txt")
            if os.path.exists(cmake_path) and not load(cmake_path).startswith("# conanws"):
                raise ConanException("Can't generate CMakeLists.txt, will overwrite existing one")
            save(cmake_path, cmake)

    def __init__(self, path, install_folder):
        self._install_folder = install_folder
        self._generator = None
        self._name = "ConanWorkspace"
        self._workspace_packages = OrderedDict()  # {reference: LocalPackage}
        if not path:
            return
        self._base_folder = os.path.dirname(path)
        content = load(path)
        self._loads(content)

    def __getitem__(self, reference):
        return self._workspace_packages.get(reference.name)

    @property
    def root(self):
        return self._root

    def _loads(self, text):
        try:
            yml = yaml.load(text)
            self._generator = yml.pop("generator", None)
            self._name = yml.pop("name", None)
            self._root = [s.strip() for s in yml.pop("root", "").split(",") if s.strip()]
            if not self._root:
                raise ConanException("Conan workspace needs at least 1 root conanfile")
            for package_name, data in yml.items():
                workspace_package = LocalPackage(self._base_folder, self._install_folder, data)
                self._workspace_packages[package_name] = workspace_package
            for package_name in self._root:
                if package_name not in self._workspace_packages:
                    raise ConanException("Root %s is not a local package" % package_name)
        except Exception as e:
            raise ConanException("There was an error parsing %s: %s" % (WORKSPACE_FILE, str(e)))



import itertools
import os
import unittest

from conans.errors import ConanException
from conans.model.ref import ConanFileReference, PackageReference
from conans.paths import EXPORT_SOURCES_TGZ_NAME, PACKAGE_TGZ_NAME
from conans.test.utils.cpp_test_files import cpp_hello_conan_files
from conans.client.tools.env import environment_append
from conans.test.utils.tools import TestClient, TestServer, NO_SETTINGS_PACKAGE_ID
from conans.util.files import save, is_dirty, gzopen_without_timestamps
from mock import mock

conanfile = """from conans import ConanFile
class MyPkg(ConanFile):
    name = "Hello0"
    version = "1.2.1"
    exports_sources = "*"

    def package(self):
        self.copy("*")
"""

conanfile_upload_query = """from conans import ConanFile
class MyPkg(ConanFile):
    name = "Hello1"
    version = "1.2.1"
    exports_sources = "*"
    settings = "os", "arch"

    def package(self):
        self.copy("*")
"""


class UploadTest(unittest.TestCase):

    def not_existing_error_test(self):
        """ Trying to upload with pattern not matched must raise an Error
        """
        client = TestClient()
        error = client.run("upload some_nonsense", ignore_error=True)
        self.assertTrue(error)
        self.assertIn("ERROR: No packages found matching pattern 'some_nonsense'",
                      client.user_io.out)

    def invalid_reference_error_test(self):
        """ Trying to upload an invalid reference must raise an Error
        """
        client = TestClient()
        error = client.run("upload some_nonsense -p hash1", ignore_error=True)
        self.assertTrue(error)
        self.assertIn("ERROR: -p parameter only allowed with a valid recipe reference",
                      client.user_io.out)

    def non_existing_recipe_error_test(self):
        """ Trying to upload a non-existing recipe must raise an Error
        """
        client = TestClient()
        error = client.run("upload Pkg/0.1@user/channel", ignore_error=True)
        self.assertTrue(error)
        self.assertIn("ERROR: There is no local conanfile exported as Pkg/0.1@user/channel",
                      client.user_io.out)

    def non_existing_package_error_test(self):
        """ Trying to upload a non-existing package must raise an Error
        """
        client = TestClient()
        error = client.run("upload Pkg/0.1@user/channel -p hash1", ignore_error=True)
        self.assertTrue(error)
        self.assertIn("ERROR: There is no local conanfile exported as Pkg/0.1@user/channel",
                      client.user_io.out)

    def _client(self):
        if not hasattr(self, "_servers"):
            servers = {}
            test_server = TestServer([("*/*@*/*", "*")], [("*/*@*/*", "*")],
                                     users={"lasote": "mypass"})
            servers["default"] = test_server
            self._servers = servers
        client = TestClient(servers=self._servers, users={"default": [("lasote", "mypass")]})
        return client

    def pattern_upload_test(self):
        client = self._client()
        client.save({"conanfile.py": conanfile})
        client.run("create . user/testing")
        client.run("upload Hello0/*@user/testing --confirm --all")
        self.assertIn("Uploading conanmanifest.txt", client.user_io.out)
        self.assertIn("Uploading conan_package.tgz", client.user_io.out)
        self.assertIn("Uploading conanfile.py", client.user_io.out)

    def query_upload_test(self):
        client = self._client()
        client.save({"conanfile.py": conanfile_upload_query})

        for os, arch in itertools.product(["Macos", "Linux", "Windows"],
                                          ["armv8", "x86_64"]):
            client.run("create . user/testing -s os=%s -s arch=%s" % (os, arch))

        # Check that the right number of packages are picked up by the queries
        client.run("upload Hello1/*@user/testing --confirm -q 'os=Windows or os=Macos'")
        for i in range(1, 5):
            self.assertIn("Uploading package %d/4" % i, client.user_io.out)
        self.assertNotIn("Package is up to date, upload skipped", client.user_io.out)

        client.run("upload Hello1/*@user/testing --confirm -q 'os=Linux and arch=x86_64'")
        self.assertIn("Uploading package 1/1", client.user_io.out)

        client.run("upload Hello1/*@user/testing --confirm -q 'arch=armv8'")
        for i in range(1, 4):
            self.assertIn("Uploading package %d/3" % i, client.user_io.out)
        self.assertIn("Package is up to date, upload skipped", client.user_io.out)

        # Check that a query not matching any packages doesn't upload any packages
        client.run("upload Hello1/*@user/testing --confirm -q 'arch=sparc'")
        self.assertNotIn("Uploading package", client.user_io.out)

        # Check that an invalid query fails
        try:
            client.run("upload Hello1/*@user/testing --confirm -q 'blah blah blah'")
        except:
            self.assertIn("Invalid package query", client.user_io.out)

    def broken_sources_tgz_test(self):
        # https://github.com/conan-io/conan/issues/2854
        client = self._client()
        client.save({"conanfile.py": conanfile,
                     "source.h": "my source"})
        client.run("create . user/testing")
        ref = ConanFileReference.loads("Hello0/1.2.1@user/testing")

        def gzopen_patched(name, mode="r", fileobj=None, compresslevel=None, **kwargs):
            raise ConanException("Error gzopen %s" % name)
        with mock.patch('conans.client.remote_manager.gzopen_without_timestamps', new=gzopen_patched):
            error = client.run("upload * --confirm", ignore_error=True)
            self.assertTrue(error)
            self.assertIn("ERROR: Error gzopen conan_sources.tgz", client.out)

            export_folder = client.client_cache.export(ref)
            tgz = os.path.join(export_folder, EXPORT_SOURCES_TGZ_NAME)
            self.assertTrue(os.path.exists(tgz))
            self.assertTrue(is_dirty(tgz))

        client.run("upload * --confirm")
        self.assertIn("WARN: Hello0/1.2.1@user/testing: Removing conan_sources.tgz, marked as dirty",
                      client.out)
        self.assertTrue(os.path.exists(tgz))
        self.assertFalse(is_dirty(tgz))

    def broken_package_tgz_test(self):
        # https://github.com/conan-io/conan/issues/2854
        client = self._client()
        client.save({"conanfile.py": conanfile,
                     "source.h": "my source"})
        client.run("create . user/testing")
        package_ref = PackageReference.loads("Hello0/1.2.1@user/testing:" +
                                             NO_SETTINGS_PACKAGE_ID)

        def gzopen_patched(name, mode="r", fileobj=None, compresslevel=None, **kwargs):
            if name == PACKAGE_TGZ_NAME:
                raise ConanException("Error gzopen %s" % name)
            return gzopen_without_timestamps(name, mode, fileobj, compresslevel, **kwargs)
        with mock.patch('conans.client.remote_manager.gzopen_without_timestamps', new=gzopen_patched):
            error = client.run("upload * --confirm --all", ignore_error=True)
            self.assertTrue(error)
            self.assertIn("ERROR: Error gzopen conan_package.tgz", client.out)

            export_folder = client.client_cache.package(package_ref)
            tgz = os.path.join(export_folder, PACKAGE_TGZ_NAME)
            self.assertTrue(os.path.exists(tgz))
            self.assertTrue(is_dirty(tgz))

        client.run("upload * --confirm --all")
        self.assertIn("WARN: Hello0/1.2.1@user/testing:%s: "
                      "Removing conan_package.tgz, marked as dirty" % NO_SETTINGS_PACKAGE_ID,
                      client.out)
        self.assertTrue(os.path.exists(tgz))
        self.assertFalse(is_dirty(tgz))

    def corrupt_upload_test(self):
        client = self._client()

        client.save({"conanfile.py": conanfile,
                     "include/hello.h": ""})
        client.run("create . frodo/stable")
        ref = ConanFileReference.loads("Hello0/1.2.1@frodo/stable")
        packages_folder = client.client_cache.packages(ref)
        pkg_id = os.listdir(packages_folder)[0]
        package_folder = os.path.join(packages_folder, pkg_id)
        save(os.path.join(package_folder, "added.txt"), "")
        os.remove(os.path.join(package_folder, "include/hello.h"))
        error = client.run("upload Hello0/1.2.1@frodo/stable --all --check", ignore_error=True)
        self.assertTrue(error)
        self.assertIn("WARN: Mismatched checksum 'added.txt'", client.user_io.out)
        self.assertIn("WARN: Mismatched checksum 'include/hello.h'", client.user_io.out)
        self.assertIn("ERROR: Cannot upload corrupted package", client.user_io.out)

    def upload_modified_recipe_test(self):
        client = self._client()

        client.save({"conanfile.py": conanfile,
                     "hello.cpp": "int i=0"})
        client.run("export . frodo/stable")
        client.run("upload Hello0/1.2.1@frodo/stable")
        self.assertIn("Uploading conanmanifest.txt", client.user_io.out)
        self.assertIn("Uploaded conan recipe 'Hello0/1.2.1@frodo/stable' to 'default'",
                      client.out)

        client2 = self._client()
        client2.save({"conanfile.py": conanfile + "\r\n#end",
                      "hello.cpp": "int i=1"})
        client2.run("export . frodo/stable")
        ref = ConanFileReference.loads("Hello0/1.2.1@frodo/stable")
        manifest = client2.client_cache.load_manifest(ref)
        manifest.time += 10
        manifest.save(client2.client_cache.export(ref))
        client2.run("upload Hello0/1.2.1@frodo/stable")
        self.assertIn("Uploading conanmanifest.txt", client2.user_io.out)
        self.assertIn("Uploaded conan recipe 'Hello0/1.2.1@frodo/stable' to 'default'",
                      client2.out)

        # first client tries to upload again
        error = client.run("upload Hello0/1.2.1@frodo/stable", ignore_error=True)
        self.assertTrue(error)
        self.assertIn("Remote recipe is newer than local recipe", client.user_io.out)
        self.assertIn("Local 'conanfile.py' using '\\n' line-ends", client.user_io.out)
        self.assertIn("Remote 'conanfile.py' using '\\r\\n' line-ends", client.user_io.out)

    def upload_unmodified_recipe_test(self):
        client = self._client()

        files = cpp_hello_conan_files("Hello0", "1.2.1", build=False)
        client.save(files)
        client.run("export . frodo/stable")
        client.run("upload Hello0/1.2.1@frodo/stable")
        self.assertIn("Uploading conanmanifest.txt", client.user_io.out)
        self.assertIn("Uploaded conan recipe 'Hello0/1.2.1@frodo/stable' to 'default'",
                      client.out)

        client2 = self._client()
        client2.save(files)
        client2.run("export . frodo/stable")
        ref = ConanFileReference.loads("Hello0/1.2.1@frodo/stable")
        manifest = client2.client_cache.load_manifest(ref)
        manifest.time += 10
        manifest.save(client2.client_cache.export(ref))
        client2.run("upload Hello0/1.2.1@frodo/stable")
        self.assertNotIn("Uploading conanmanifest.txt", client2.out)
        self.assertNotIn("Uploaded conan recipe 'Hello0/1.2.1@frodo/stable' to 'default'",
                         client2.out)
        self.assertIn("Recipe is up to date, upload skipped", client2.out)

        # first client tries to upload again
        client.run("upload Hello0/1.2.1@frodo/stable")
        self.assertNotIn("Uploading conanmanifest.txt", client.out)
        self.assertNotIn("Uploaded conan recipe 'Hello0/1.2.1@frodo/stable' to 'default'",
                         client.out)
        self.assertIn("Recipe is up to date, upload skipped", client.out)

    def upload_unmodified_package_test(self):
        client = self._client()

        client.save({"conanfile.py": conanfile,
                     "hello.cpp": ""})
        client.run("create . frodo/stable")
        client.run("upload Hello0/1.2.1@frodo/stable --all")

        client2 = self._client()
        client2.save({"conanfile.py": conanfile,
                      "hello.cpp": ""})
        client2.run("create . frodo/stable")
        client2.run("upload Hello0/1.2.1@frodo/stable --all")
        self.assertIn("Recipe is up to date, upload skipped", client2.out)
        self.assertNotIn("Uploading conanfile.py", client2.out)
        self.assertNotIn("Uploading conan_sources.tgz", client2.out)
        self.assertNotIn("Uploaded conan recipe 'Hello0/1.2.1@frodo/stable' to 'default'",
                         client2.out)
        self.assertNotIn("Uploading conaninfo.txt", client2.out)  # conaninfo NOT changed
        self.assertNotIn("Uploading conan_package.tgz", client2.out)
        self.assertIn("Package is up to date, upload skipped", client2.out)

        # first client tries to upload again
        client.run("upload Hello0/1.2.1@frodo/stable --all")
        self.assertIn("Recipe is up to date, upload skipped", client.out)
        self.assertNotIn("Uploading conanfile.py", client.out)
        self.assertNotIn("Uploading conan_sources.tgz", client.out)
        self.assertNotIn("Uploaded conan recipe 'Hello0/1.2.1@frodo/stable' to 'default'",
                         client.out)
        self.assertNotIn("Uploading conaninfo.txt", client.out)  # conaninfo NOT changed
        self.assertNotIn("Uploading conan_package.tgz", client2.out)
        self.assertIn("Package is up to date, upload skipped", client2.out)

    def no_overwrite_argument_collision_test(self):
        client = self._client()
        client.save({"conanfile.py": conanfile,
                     "hello.cpp": ""})
        client.run("create . frodo/stable")

        # Not valid values as arguments
        error = client.run("upload Hello0/1.2.1@frodo/stable --no-overwrite kk", ignore_error=True)
        self.assertTrue(error)
        self.assertIn("ERROR", client.out)

        # --force not valid with --no-overwrite
        error = client.run("upload Hello0/1.2.1@frodo/stable --no-overwrite --force",
                           ignore_error=True)
        self.assertTrue(error)
        self.assertIn("ERROR: '--no-overwrite' argument cannot be used together with '--force'",
                      client.out)

    def upload_no_overwrite_all_test(self):
        conanfile_new = """from conans import ConanFile, tools
class MyPkg(ConanFile):
    name = "Hello0"
    version = "1.2.1"
    exports_sources = "*"
    options = {"shared": [True, False]}
    default_options = "shared=False"

    def build(self):
        if tools.get_env("MY_VAR", False):
            open("file.h", 'w').close()

    def package(self):
        self.copy("*.h")
"""
        client = self._client()
        client.save({"conanfile.py": conanfile_new,
                     "hello.h": "",
                     "hello.cpp": ""})
        client.run("create . frodo/stable")

        # First time upload
        client.run("upload Hello0/1.2.1@frodo/stable --all --no-overwrite")
        self.assertNotIn("Forbidden overwrite", client.out)
        self.assertIn("Uploading Hello0/1.2.1@frodo/stable", client.out)

        # CASE: Upload again
        client.run("upload Hello0/1.2.1@frodo/stable --all --no-overwrite")
        self.assertIn("Recipe is up to date, upload skipped", client.out)
        self.assertIn("Package is up to date, upload skipped", client.out)
        self.assertNotIn("Forbidden overwrite", client.out)

        # CASE: Without changes
        client.run("create . frodo/stable")
        # upload recipe and packages
        client.run("upload Hello0/1.2.1@frodo/stable --all --no-overwrite")
        self.assertIn("Recipe is up to date, upload skipped", client.all_output)
        self.assertIn("Package is up to date, upload skipped", client.out)
        self.assertNotIn("Forbidden overwrite", client.out)

        # CASE: When recipe and package changes
        new_recipe = conanfile_new.replace("self.copy(\"*.h\")",
                                           "self.copy(\"*.h\")\n        self.copy(\"*.cpp\")")
        client.save({"conanfile.py": new_recipe})
        client.run("create . frodo/stable")
        # upload recipe and packages
        # *1
        error = client.run("upload Hello0/1.2.1@frodo/stable --all --no-overwrite",
                           ignore_error=True)
        if not client.revisions:  # The --no-overwrite makes no sense with revisions
            self.assertTrue(error)
            self.assertIn("Forbidden overwrite", client.out)
            self.assertNotIn("Uploading conan_package.tgz", client.out)

        # CASE: When package changes
        client.run("upload Hello0/1.2.1@frodo/stable --all")
        with environment_append({"MY_VAR": "True"}):
            client.run("create . frodo/stable")
        # upload recipe and packages
        error = client.run("upload Hello0/1.2.1@frodo/stable --all --no-overwrite",
                           ignore_error=True)
        if not client.revisions:
            self.assertTrue(error)
            self.assertIn("Recipe is up to date, upload skipped", client.out)
            self.assertIn("ERROR: Local package is different from the remote package", client.out)
            self.assertIn("Forbidden overwrite", client.out)
            self.assertNotIn("Uploading conan_package.tgz", client.out)
        else:
            self.assertIn("Uploading conan_package.tgz", client.out)

    def upload_no_overwrite_recipe_test(self):
        conanfile_new = """from conans import ConanFile, tools
class MyPkg(ConanFile):
    name = "Hello0"
    version = "1.2.1"
    exports_sources = "*"
    options = {"shared": [True, False]}
    default_options = "shared=False"

    def build(self):
        if tools.get_env("MY_VAR", False):
            open("file.h", 'w').close()

    def package(self):
        self.copy("*.h")
"""
        client = self._client()
        client.save({"conanfile.py": conanfile_new,
                     "hello.h": "",
                     "hello.cpp": ""})
        client.run("create . frodo/stable")

        # First time upload
        client.run("upload Hello0/1.2.1@frodo/stable --all --no-overwrite recipe")
        self.assertNotIn("Forbidden overwrite", client.out)
        self.assertIn("Uploading Hello0/1.2.1@frodo/stable", client.out)

        # Upload again
        client.run("upload Hello0/1.2.1@frodo/stable --all --no-overwrite recipe")
        self.assertIn("Recipe is up to date, upload skipped", client.out)
        self.assertIn("Package is up to date, upload skipped", client.out)
        self.assertNotIn("Forbidden overwrite", client.out)

        # Create without changes
        # *1
        client.run("create . frodo/stable")
        client.run("upload Hello0/1.2.1@frodo/stable --all --no-overwrite recipe")
        self.assertIn("Recipe is up to date, upload skipped", client.out)
        self.assertIn("Package is up to date, upload skipped", client.out)
        self.assertNotIn("Forbidden overwrite", client.out)

        # Create with recipe and package changes
        new_recipe = conanfile_new.replace("self.copy(\"*.h\")",
                                           "self.copy(\"*.h\")\n        self.copy(\"*.cpp\")")
        client.save({"conanfile.py": new_recipe})
        client.run("create . frodo/stable")
        # upload recipe and packages
        error = client.run("upload Hello0/1.2.1@frodo/stable --all --no-overwrite recipe",
                           ignore_error=True)
        if not client.revisions:
            self.assertTrue(error)
            self.assertIn("Forbidden overwrite", client.out)
            self.assertNotIn("Uploading package", client.out)

            # Create with package changes
            client.run("upload Hello0/1.2.1@frodo/stable --all")
            with environment_append({"MY_VAR": "True"}):
                client.run("create . frodo/stable")
            # upload recipe and packages
            client.run("upload Hello0/1.2.1@frodo/stable --all --no-overwrite recipe")
            self.assertIn("Recipe is up to date, upload skipped", client.out)
            self.assertIn("Uploading conan_package.tgz", client.out)
            self.assertNotIn("Forbidden overwrite", client.out)
        else:
            self.assertIn("Uploading conan_package.tgz", client.out)

    def skip_upload_test(self):
        """ Check that the option --dry does not upload anything
        """
        client = self._client()

        files = cpp_hello_conan_files("Hello0", "1.2.1", build=False)
        client.save(files)
        client.run("export . frodo/stable")
        client.run("install Hello0/1.2.1@frodo/stable --build=missing")
        client.run("upload Hello0/1.2.1@frodo/stable -r default --all --skip-upload")

        # dry run should not upload
        self.assertNotIn("Uploading conan_package.tgz", client.user_io.out)

        # but dry run should compress
        self.assertIn("Compressing recipe...", client.user_io.out)
        self.assertIn("Compressing package...", client.user_io.out)

        client.run("search -r default")
        # after dry run nothing should be on the server ...
        self.assertNotIn("Hello0/1.2.1@frodo/stable", client.user_io.out)

        # now upload, the stuff should NOT be recompressed
        client.run("upload Hello0/1.2.1@frodo/stable -r default --all")

        # check for upload message
        self.assertIn("Uploading conan_package.tgz", client.user_io.out)

        # check if compressed files are re-used
        self.assertNotIn("Compressing recipe...", client.user_io.out)
        self.assertNotIn("Compressing package...", client.user_io.out)

        # now it should be on the server
        client.run("search -r default")
        self.assertIn("Hello0/1.2.1@frodo/stable", client.user_io.out)

    def upload_without_sources_test(self):
        client = self._client()
        conanfile = """from conans import ConanFile
class Pkg(ConanFile):
    pass
"""
        client.save({"conanfile.py": conanfile})
        client.run("create . Pkg/0.1@user/testing")
        client.run("upload * --all --confirm")
        client2 = self._client()
        client2.run("install Pkg/0.1@user/testing")
        client2.run("remote remove default")
        server2 = TestServer([("*/*@*/*", "*")], [("*/*@*/*", "*")],
                             users={"lasote": "mypass"})
        client2.users = {"server2": [("lasote", "mypass")]}
        client2.servers = {"server2": server2}
        client2.update_servers()
        client2.run("upload * --all --confirm -r=server2")
        self.assertIn("Uploading conanfile.py", client2.out)
        self.assertIn("Uploading conan_package.tgz", client2.out)

    def upload_login_prompt_disabled_no_user_test(self):
        """ Without user info, uploads should fail when login prompt has been disabled.
        """
        files = cpp_hello_conan_files("Hello0", "1.2.1", build=False)
        client = self._client()
        client.save(files)
        client.run("config set general.non_interactive=True")
        client.run("create . user/testing")
        client.run("user -c")
        error = client.run("upload Hello0/1.2.1@user/testing", ignore_error=True)
        self.assertTrue(error)
        self.assertIn('ERROR: Conan interactive mode disabled', client.out)
        self.assertNotIn("Uploading conanmanifest.txt", client.out)
        self.assertNotIn("Uploading conanfile.py", client.out)
        self.assertNotIn("Uploading conan_export.tgz", client.out)

    def upload_login_prompt_disabled_user_not_authenticated_test(self):
        """ When a user is not authenticated, uploads should fail when login prompt has been disabled.
        """
        files = cpp_hello_conan_files("Hello0", "1.2.1", build=False)
        client = self._client()
        client.save(files)
        client.run("config set general.non_interactive=True")
        client.run("create . user/testing")
        client.run("user -c")
        client.run("user lasote")
        error = client.run("upload Hello0/1.2.1@user/testing", ignore_error=True)
        self.assertTrue(error)
        self.assertIn('ERROR: Conan interactive mode disabled', client.out)
        self.assertNotIn("Uploading conanmanifest.txt", client.out)
        self.assertNotIn("Uploading conanfile.py", client.out)
        self.assertNotIn("Uploading conan_export.tgz", client.out)
        self.assertNotIn("Please enter a password for \"lasote\" account:", client.out)

    def upload_login_prompt_disabled_user_authenticated_test(self):
        """ When a user is authenticated, uploads should work even when login prompt has been disabled.
        """
        files = cpp_hello_conan_files("Hello0", "1.2.1", build=False)
        client = self._client()
        client.save(files)
        client.run("config set general.non_interactive=True")
        client.run("create . user/testing")
        client.run("user -c")
        client.run("user lasote -p mypass")
        client.run("upload Hello0/1.2.1@user/testing")
        self.assertIn("Uploading conanmanifest.txt", client.out)
        self.assertIn("Uploading conanfile.py", client.out)
        self.assertIn("Uploading conan_export.tgz", client.out)



import os
from conans.client.cmd.test import PackageTester
from conans.errors import ConanException


def get_test_conanfile_path(tf, conanfile_path):
    """Searches in the declared test_folder or in the standard locations"""

    if tf is False:
        # Look up for testing conanfile can be disabled if tf (test folder) is False
        return None

    test_folders = [tf] if tf else ["test_package", "test"]
    base_folder = os.path.dirname(conanfile_path)
    for test_folder_name in test_folders:
        test_folder = os.path.join(base_folder, test_folder_name)
        test_conanfile_path = os.path.join(test_folder, "conanfile.py")
        if os.path.exists(test_conanfile_path):
            return test_conanfile_path
    else:
        if tf:
            raise ConanException("test folder '%s' not available, or it doesn't have a conanfile.py"
                                 % tf)


def create(reference, manager, user_io, profile, remote_name, update, build_modes, manifest_folder,
           manifest_verify, manifest_interactive, keep_build, test_build_folder, test_folder,
           conanfile_path):

    test_conanfile_path = get_test_conanfile_path(test_folder, conanfile_path)

    if test_conanfile_path:
        pt = PackageTester(manager, user_io)
        pt.install_build_and_test(test_conanfile_path, reference, profile, remote_name, update,
                                  build_modes=build_modes,
                                  manifest_folder=manifest_folder,
                                  manifest_verify=manifest_verify,
                                  manifest_interactive=manifest_interactive,
                                  keep_build=keep_build,
                                  test_build_folder=test_build_folder)
    else:
        manager.install(reference=reference,
                        create_reference=reference,
                        install_folder=None,  # Not output anything
                        manifest_folder=manifest_folder,
                        manifest_verify=manifest_verify,
                        manifest_interactive=manifest_interactive,
                        remote_name=remote_name,
                        profile=profile,
                        build_modes=build_modes,
                        update=update,
                        keep_build=keep_build)



import copy
import os

from conans.client.build.compiler_flags import build_type_define, build_type_flags, visual_runtime, format_defines, \
    include_path_option, parallel_compiler_cl_flag
from conans.client.build.cppstd_flags import cppstd_flag


class VisualStudioBuildEnvironment(object):
    """
    - LIB: library paths with semicolon separator
    - CL: /I (include paths)
    - _LINK_: linker options and libraries

    https://msdn.microsoft.com/en-us/library/19z1t1wy.aspx
    https://msdn.microsoft.com/en-us/library/fwkeyyhe.aspx
    https://msdn.microsoft.com/en-us/library/9s7c9wdw.aspx
    https://msdn.microsoft.com/en-us/library/6y6t9esh.aspx

    """
    def __init__(self, conanfile, with_build_type_flags=True):
        """
        :param conanfile: ConanFile instance
        :param quote_paths: The path directories will be quoted. If you are using the vars together with
                            environment_append keep it to True, for virtualbuildenv quote_paths=False is required.
        """
        self._with_build_type_flags = with_build_type_flags

        self._conanfile = conanfile
        self._settings = conanfile.settings
        self._deps_cpp_info = conanfile.deps_cpp_info
        self._runtime = self._settings.get_safe("compiler.runtime")

        self.include_paths = conanfile.deps_cpp_info.include_paths
        self.lib_paths = conanfile.deps_cpp_info.lib_paths
        self.defines = copy.copy(conanfile.deps_cpp_info.defines)
        self.flags = self._configure_flags()
        self.cxx_flags = copy.copy(self._deps_cpp_info.cppflags)
        self.link_flags = self._configure_link_flags()
        self.libs = conanfile.deps_cpp_info.libs
        self.std = self._std_cpp()
        self.parallel = False

    def _configure_link_flags(self):
        ret = copy.copy(self._deps_cpp_info.exelinkflags)
        ret.extend(self._deps_cpp_info.sharedlinkflags)
        return ret

    def _configure_flags(self):
        ret = copy.copy(self._deps_cpp_info.cflags)
        ret.extend(vs_build_type_flags(self._settings, with_flags=self._with_build_type_flags))
        return ret

    def _get_cl_list(self, quotes=True):
        # FIXME: It should be managed with the compiler_flags module
        # But need further investigation about the quotes and so on, so better to not break anything
        if quotes:
            ret = ['%s"%s"' % (include_path_option, lib) for lib in self.include_paths]
        else:
            ret = ['%s%s' % (include_path_option, lib) for lib in self.include_paths]

        runtime = visual_runtime(self._runtime)
        if runtime:
            ret.append(runtime)

        ret.extend(format_defines(self.defines))
        ret.extend(self.flags)
        ret.extend(self.cxx_flags)

        if self.parallel:  # Build source in parallel
            ret.append(parallel_compiler_cl_flag(output=self._conanfile.output))

        if self.std:
            ret.append(self.std)

        return ret

    def _get_link_list(self):
        def format_lib(lib):
            return lib if lib.endswith('.lib') else '%s.lib' % lib

        ret = [flag for flag in self.link_flags]  # copy
        ret.extend([format_lib(lib) for lib in self.libs])

        return ret

    @property
    def vars(self):
        """Used in conanfile with environment_append"""
        flags = self._get_cl_list()
        link_flags = self._get_link_list()

        cl_args = " ".join(flags) + _environ_value_prefix("CL")
        link_args = " ".join(link_flags)
        lib_paths = ";".join(['%s' % lib for lib in self.lib_paths]) + _environ_value_prefix("LIB", ";")
        return {"CL": cl_args,
                "LIB": lib_paths,
                "_LINK_": link_args}

    @property
    def vars_dict(self):
        """Used in virtualbuildenvironment"""
        # Here we do not quote the include paths, it's going to be used by virtual environment
        cl = self._get_cl_list(quotes=False)
        link = self._get_link_list()

        lib = [lib for lib in self.lib_paths]  # copy

        if os.environ.get("CL", None):
            cl.append(os.environ.get("CL"))

        if os.environ.get("LIB", None):
            lib.append(os.environ.get("LIB"))

        if os.environ.get("_LINK_", None):
            link.append(os.environ.get("_LINK_"))

        ret = {"CL": cl,
               "LIB": lib,
               "_LINK_": link}
        return ret

    def _std_cpp(self):
        return vs_std_cpp(self._settings)


def vs_build_type_flags(settings, with_flags=True):
    build_type = settings.get_safe("build_type")
    ret = []
    btd = build_type_define(build_type=build_type)
    if btd:
        ret.extend(format_defines([btd]))
    if with_flags:
        # When using to build a vs project we don't want to adjust these flags
        btfs = build_type_flags("Visual Studio", build_type=build_type,
                                vs_toolset=settings.get_safe("compiler.toolset"))
        if btfs:
            ret.extend(btfs)

    return ret


def vs_std_cpp(settings):
    if settings.get_safe("compiler") == "Visual Studio" and \
       settings.get_safe("cppstd"):
        flag = cppstd_flag(settings.get_safe("compiler"),
                           settings.get_safe("compiler.version"),
                           settings.get_safe("cppstd"))
        return flag
    return None


def _environ_value_prefix(var_name, prefix=" "):
    if os.environ.get(var_name, ""):
        return "%s%s" % (prefix, os.environ.get(var_name, ""))
    else:
        return ""
